# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Compute cross coherence between all nodes in a time series.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>
.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""

import numpy
import matplotlib.mlab as mlab
from matplotlib.pylab import detrend_linear
import tvb.datatypes.spectral as spectral
from tvb.basic.logger.builder import get_logger
from tvb.basic.neotraits.api import narray_describe

log = get_logger(__name__)


# TODO: Should do this properly, ie not with mlab, returning both coherence and
#      the complex coherence spectra, then supporting magnitude squared
#      coherence, etc in a similar fashion to the FourierSpectrum datatype...


def _hamming(M, sym=True):
    """
    The M-point Hamming window.
    From scipy.signal
    """
    if M < 1:
        return numpy.array([])
    if M == 1:
        return numpy.ones(1, 'd')
    odd = M % 2
    if not sym and not odd:
        M = M + 1
    n = numpy.arange(0, M)
    w = 0.54 - 0.46 * numpy.cos(2.0 * numpy.pi * n / (M - 1))
    if not sym and not odd:
        w = w[:-1]
    return w


def coherence_mlab(data, sample_rate, nfft=256):
    _, nsvar, nnode, nmode = data.shape
    # (frequency, nodes, nodes, state-variables, modes)
    coh_shape = nfft/2 + 1, nnode, nnode, nsvar, nmode
    coh = numpy.zeros(coh_shape)
    for mode in range(nmode):
        for var in range(nsvar):
            data = data[:, var, :, mode].copy()
            data -= data.mean(axis=0)[numpy.newaxis, :]
            for n1 in range(nnode):
                for n2 in range(nnode):
                    cxy, freq = mlab.cohere(data[:, n1], data[:, n2],
                                            NFFT=nfft,
                                            Fs=sample_rate,
                                            detrend=detrend_linear,
                                            window=mlab.window_none)
                    coh[:, n1, n2, var, mode] = cxy
    return coh, freq


def _coherence(data, sample_rate, nfft=256, imag=False):
    "Vectorized coherence calculation by windowed FFT"
    nt, ns, nn, nm = data.shape
    nwin = nt // nfft
    if nwin < 1:
        raise ValueError(
            "Not enough time points ({0}) to compute an FFT, given a "
            "window size of nfft={1}.".format(nt, nfft))
    # ignore leftover data; need shape (nn, ... , nwin, nfft)
    wins = data[:int(nwin * nfft)]\
        .copy()\
        .transpose((2, 1, 3, 0))\
        .reshape((nn, ns, nm, nwin, nfft))
    wins *= _hamming(nfft)
    F = numpy.fft.fft(wins)
    fs = numpy.fft.fftfreq(nfft, 1e3 / sample_rate)
    # broadcasts to [node_i, node_j, ..., window, time]
    G = F[:, numpy.newaxis] * F.conj()
    if imag:
        G = G.imag
    dG = numpy.array([G[i, i] for i in range(nn)])
    C = (numpy.abs(G)**2 / (dG[:, numpy.newaxis] * dG)).mean(axis=-2)
    mask = fs > 0.0
    # C_ = numpy.abs(C.mean(axis=0).mean(axis=0))
    return numpy.transpose(C[..., mask], (4, 0, 1, 2, 3)), fs[mask]


def calculate_cross_coherence(time_series, nfft):
    """
    # type: (TimeSeries, int)  -> CoherenceSpectrum
    # Adapter for cross-coherence algorithm(s)
    # Evaluate coherence on time series.

    Parameters
    __________
    time_series : TimeSeries
    The TimeSeries to which the Cross Coherence is to be applied.

    nfft : int
    Data-points per block (should be a power of 2).
    """

    srate = time_series.sample_rate
    coh, freq = _coherence(time_series.data, srate, nfft=nfft)
    log.debug("coherence")
    log.debug(narray_describe(coh))
    log.debug("freq")
    log.debug(narray_describe(freq))

    spec = spectral.CoherenceSpectrum(
        source=time_series,
        nfft=nfft,
        array_data=coh.astype(numpy.float64),
        frequency=freq)
    return spec




# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

import abc
import types
import uuid
import numpy
import numpy as np
import pytest
from tvb.datatypes import connectivity
from tvb.simulator import (simulator, models, coupling, integrators, monitors, noise)
from tvb.basic.neotraits._core import TraitProperty
from tvb.basic.neotraits.api import (
    HasTraits, Attr, NArray, Final, List, trait_property, narray_summary_info, narray_describe,
    Int, Float, Range, cached_trait_property, LinspaceRange, Dim
)
from tvb.basic.neotraits.ex import TraitTypeError, TraitValueError, TraitAttributeError, TraitError


def test_simple_declaration():
    class A(HasTraits):
        fi = Attr(int)


def test_simple_instantiation():
    class A(HasTraits):
        fi = Attr(int)

    a2 = A(fi=12)

    with pytest.raises(TraitError):
        a3 = A(fi_typo=43)


def test_types_enforced():
    class A(HasTraits):
        fi = Attr(int, default=0)

    ai = A()

    with pytest.raises(TypeError):
        ai.fi = 'ana'


def test_defaults_and_required():
    class A(HasTraits):
        fi = Attr(int, default=3)
        ra = Attr(str)
        vi = Attr(str, required=False)

    aok = A(ra='opaline')
    aok.vi = None

    aincomplete = A()

    with pytest.raises(ValueError):
        a = A(ra='opaline')
        a.ra = None

    with pytest.raises(TraitAttributeError):
        aincomplete.ra

    with pytest.raises(TraitAttributeError):
        aincomplete.validate()

    # the default configure calls validate
    with pytest.raises(TraitAttributeError):
        aincomplete.configure()


def test_default_compatible_with_field_type_and_choices():
    with pytest.raises(TraitError):
        class A(HasTraits):
            fi = Attr(int, choices=(3, 4), default=2345)

    with pytest.raises(TraitError):
        class A(HasTraits):
            fi = Attr(int, default=23.45)

    with pytest.raises(TraitError):
        class A(HasTraits):
            fi = Attr(int, choices=(1.2, 4.2))

        a = A()
        a.fi = 1.2  # the error is late, but better than never


def test_postinitchecks():
    class A(HasTraits):
        is_foo = Attr(bool, default=False, doc='is it a foo?')
        name = Attr(str)

        def __init__(self, **kwargs):
            super(A, self).__init__(**kwargs)
            self.name = 'opaline'

    aok = A()


def test_composition():
    class A(HasTraits):
        fi = Attr(int, default=3)
        ra = Attr(str, default='ela')

    class B(A):
        opt = Attr(str, required=False)

    class Comp(HasTraits):
        a = Attr(A)
        b = Attr(B)

    comp = Comp(a=A(), b=B())


def test_inheritance():
    class A(HasTraits):
        a = Attr(str)

    class B(HasTraits):
        b = Attr(int)
        c = Attr(float)

    # mixin should not support Attr
    class Mixin(object):
        m = Attr(int, default=42)

    class Inherit(Mixin, B, A):
        c = Attr(str)

    # c has been correctly overridden and is of type str not float
    t = Inherit(a="ana", c="are", b=2)

    assert t.c == "are"

    # we don't support Attr in mixins
    # For simplicity of implementation it fails quite late at use time not when Inherit is declared
    with pytest.raises(AttributeError):
        t.m = 2


def test_attr_misuse_fails():
    with pytest.raises(TypeError):
        class A(HasTraits):
            a = Attr('ana')


def test_declarative_attrs():
    class A(HasTraits):
        a = Attr(str)

    class B(HasTraits):
        b = Attr(int)
        c = Attr(float)

    class Inherit(B, A):
        c = Attr(str)

    assert set(Inherit.declarative_attrs) == {'gid', 'c', 'b', 'a'}
    assert Inherit._own_declarative_attrs == ('c',)
    assert Inherit.own_declarative_attrs == ('c',)

    t = Inherit(a="ana", c="are", b=2)

    # Notice that declarative_attrs is not available on the instances only on the class.
    # This is a bit unusual, but typical for metamethods
    with pytest.raises(AttributeError):
        t.declarative_attrs

    assert set(type(t).declarative_attrs) == {'gid', 'c', 'b', 'a'}


def test_mutable_defaults():
    class A(HasTraits):
        mut = Attr(list, default=[3, 4])
        mut2 = Attr(list, default=[42])

        def __init__(self, **kwargs):
            super(A, self).__init__(**kwargs)
            # copy the default in the constructor
            # and avoid the common shared default
            self.mut = self.mut[:]

    aok = A()

    assert aok.mut is not A.mut.default
    assert aok.mut2 is A.mut2.default

    class B(HasTraits):
        # a better option is to use a lambda default
        mut_better = Attr(list, default=lambda: [3, 4])

    bok = B()

    assert bok.mut_better is not B.mut_better.default
    assert aok.mut2 is A.mut2.default


def test_mutable_unassigned_defaults():
    class B(HasTraits):
        mut_shared = Attr(list, default=[3, 4])
        mut_better = Attr(list, default=lambda: [3, 4])

    bok = B()
    bok2 = B()
    bok.mut_better.append(42)
    assert bok.mut_better == [3, 4, 42]
    assert bok2.mut_better == [3, 4]

    bok.mut_shared.append(42)
    assert bok.mut_shared == [3, 4, 42]
    assert bok2.mut_shared == [3, 4, 42]


def test_late_attr_binding_fail():
    class F(HasTraits):
        f = Attr(str)

    F.newdynamic = Attr(str)
    del F.f


def test_mro_fail():
    class A(HasTraits): pass

    class B(A): pass

    with pytest.raises(TypeError) as ex:
        class Inherit(A, B):
            """
            simple mro contradiction. rule 1 Inherit(A, B) => A < B in mro ; rule 2 B subclass A => B < A
            """

        assert 'consistent method resolution' in str(ex.value)


def test_narr_simple():
    class Boo(HasTraits):
        x = NArray(shape=(Dim.any, Dim.any), dtype=np.dtype(np.int64))

    boo = Boo(x=np.array([[1, 4]]))
    boo.x = np.array([[1], [2]])


def test_narr_enforcing():
    with pytest.raises(TypeError):
        class Boo(HasTraits):
            x = NArray(dtype=np.dtype(np.int64), default=np.eye(2))

    with pytest.raises(ValueError):
        # bad ndim default
        class Boo(HasTraits):
            x = NArray(shape=(Dim.any, Dim.any, Dim.any), default=np.linspace(1, 3, 10))

    with pytest.raises(TypeError):
        # default should be ndarray
        class Boo(HasTraits):
            x = NArray(default=[1, 2, 4])

    class Boo(HasTraits):
        x = NArray(shape=(Dim.any, Dim.any), default=np.eye(2))
        y = NArray(dtype=np.dtype(int), default=np.arange(12), domain=list(range(5)))

    # only the defaults are checked for domain compliance
    boo = Boo(y=np.arange(10))
    boo.x = np.eye(5)
    boo.x = np.eye(2, dtype=float)
    # only the defaults are checked for domain compliance
    boo.y = np.arange(12)


def test_narr_accepts_safely_convertible_values():
    class Boo(HasTraits):
        x = NArray(dtype=np.float64)
        y = NArray(dtype=np.float32)

    boo = Boo()
    # float32 -> float64 is safe
    boo.x = numpy.array([2.0, 4.3], dtype=np.float32)
    # destination dtype is kept, assigned value is converted, maybe copied
    assert boo.x.dtype == np.float64
    # int -> float64 is safe
    boo.x = numpy.arange(5)
    # however int -> float32 is not
    with pytest.raises(TraitError):
        boo.y = numpy.arange(5)


def test_choices():
    class A(HasTraits):
        x = Attr(str, default='ana', choices=('ana', 'are', 'mere'))
        cx = Final('ana')

    a = A()

    with pytest.raises(ValueError):
        a.x = 'hell'


def test_const_attr():
    class A(HasTraits):
        cx = Final('ana')
        final_c = Attr(int, final=True)

    a = A()
    # in neotraits uninitialized required attrs with no defaults raise on read
    with pytest.raises(TraitAttributeError):
        a.final_c

    a.final_c = 42

    with pytest.raises(TraitAttributeError):
        a.cx = 'hell'

    with pytest.raises(TraitAttributeError):
        a.final_c = 'hell'

    # revealing the insides.
    a.__dict__['final_c'] = None
    a.final_c = 24


def test_named_dimensions():
    class A(HasTraits):
        x = NArray(dim_names=('time', 'space'), required=False)

    assert A.x.ndim == 2
    a = A()
    a.x = np.eye(4)
    assert a.x.shape == (4, 4)
    assert type(a).x.dim_names[0] == 'time'

    with pytest.raises(TraitValueError):
        # ndim dim_names contradiction
        NArray(dim_names=('time', 'space'), shape=(Dim.any, Dim.any, Dim.any, Dim.any))


def test_ndim_enforced():
    class A(HasTraits):
        x = NArray(shape=(Dim.any, Dim.any))

    a = A()
    with pytest.raises(TraitValueError):
        a.x = numpy.arange(4)


def test_dims():
    class A(HasTraits):
        n_nodes = Dim(doc='number of nodes')
        n_sv = Dim()
        w = NArray(shape=(n_nodes, n_nodes))

    class B(A):
        tsr = NArray(shape=(Dim.any, A.n_nodes, A.n_sv))

    a = A(n_sv=2)

    with pytest.raises(TraitAttributeError):
        # dims are required, so reading a non-initialized one fails
        a.n_nodes

    a.n_nodes = 42
    assert a.n_sv == 2

    b = B(n_sv=2)

    with pytest.raises(TraitAttributeError):
        # w references the undefined n_nodes
        b.w = numpy.eye(2)


def test_lists():
    class A(HasTraits):
        picked_dimensions = List(of=str, default=('time', 'space'), choices=('time', 'space', 'measurement'))

    a = A()

    with pytest.raises(ValueError):
        a.picked_dimensions = ['value not in choices']

    # on set we can complain about types and wrong choices
    with pytest.raises(TypeError):
        a.picked_dimensions = [76]

    a.picked_dimensions = ['time', 'space']
    # however there is little we can do against mutation
    a.picked_dimensions.append(76)

    # unless you assign a tuple
    with pytest.raises(AttributeError):
        a.picked_dimensions = ('time', 'space')
        a.picked_dimensions.append(76)


def test_list_default_right_type():
    with pytest.raises(TypeError):
        class A(HasTraits):
            picked_dimensions = List(of=str, default=('time', 42.24))


def test_list_default_must_respect_choices():
    with pytest.raises(ValueError):
        class A(HasTraits):
            picked_dimensions = List(
                of=str,
                default=('time',),
                choices=('a', 'b', 'c')
            )


def test_str_ndarrays():
    class A(HasTraits):
        s = NArray(dtype='U5')

    a = A(s=np.array(['ana', 'a', 'adus', 'mere']))
    # but users will expect python list[str] like behaviour and then this happens
    a.s[0] = 'Georgiana'
    assert 'Georgiana' != a.s[0]
    assert 'Georg' == a.s[0]

    # dtype(str) used to be dtype('|S0'), not any more
    with pytest.raises(TypeError):
        class A(HasTraits):
            s = NArray(dtype='|S0', default=np.array(['eli']))
        # fails because the declared type |S0 is different from |S3
        # it is not only different but not compatible


def test_reusing_attribute_instances_fail():
    with pytest.raises(AttributeError):
        class A(HasTraits):
            a, b = [Attr(int)] * 2


def test_changing_Attr_after_bind_fails():
    class A(HasTraits):
        a = Attr(str)

    with pytest.raises(AttributeError):
        A.a.default = 'ana'

    with pytest.raises(AttributeError):
        del A.a.default


def test_declarative_property():
    class A(HasTraits):
        x = NArray(
            label='th x',
            doc='the mighty x'
        )

        def __init__(self, **kwargs):
            super(A, self).__init__(**kwargs)
            self.x3_calls = 0
            self.expensive_once_calls = 0

        def tw(self):
            return self.x * 2

        x2 = TraitProperty(tw, NArray(label='x * 2'))

        @trait_property(NArray(label='3 * x', doc='use the docstring, it is nicer'))
        def x3(self):
            """
            this will be added to the doc of the NArray
            Encouraged place for doc
            """
            self.x3_calls += 1
            return self.x * 3

        @cached_trait_property(NArray(label='expensive'))
        def expensive_once(self):
            self.expensive_once_calls += 1
            d = self.x[:, numpy.newaxis]
            return d + d.T

    a = A()
    a.x = np.arange(12)
    assert set(A.declarative_props) == {'x3', 'x2', 'expensive_once'}
    assert (a.x2 == a.x * 2).all()
    assert (a.x3 == a.x * 3).all()

    a.x3
    assert a.x3_calls == 2
    a.expensive_once
    assert a.expensive_once_calls == 1
    # this does not call expensive_once again
    assert a.expensive_once.shape == (12, 12)
    assert a.expensive_once_calls == 1
    a.expensive_once = numpy.eye(4, dtype=float)
    assert a.expensive_once.shape == (4, 4)
    # invalidate cache by removing the instance state
    del a.expensive_once
    a.expensive_once
    assert a.expensive_once_calls == 2
    assert a.expensive_once.shape == (12, 12)


def test_bool_failure():
    class A(HasTraits):
        does_it_sink = Attr(bool)

    a = A()

    with pytest.raises(TraitTypeError) as e:
        a.does_it_sink = (numpy.zeros(4) == 0).all()

    # bool is not numpy.bool_ !!
    # should we introduce a Bool Attr like we had to do for the Int ??

    # I'd blame numpy if the following assertion would have failed:
    # WHY python? Why must bool be a int, this is not C and even there ...
    assert True + True == 2
    # for more fun philosophy:
    # truth without truth is falsity
    assert True - True == False
    # Half a truth is a falsity
    assert True // 2 == False


def test_int_attribute():
    class A(HasTraits):
        a = Int()
        b = Int(field_type=np.int8)
        c = Int(field_type=np.uint16)

    ainst = A()
    assert ainst.b == 0

    # type is out of bounds but value is within the bounds. So this is ok
    ainst.b = int(42)
    # values are not only checked for compatibility but converted to the declared type
    assert type(ainst.b) == np.int8
    ainst.b = np.int64(4)

    with pytest.raises(TypeError):
        # out of bounds for a int8
        ainst.b = 102345

    with pytest.raises(TypeError):
        # floats can't be safely cast to int
        ainst.a = 3.12

    with pytest.raises(TypeError):
        # floats are not ok even when they don't have decimals
        ainst.a = 4.0

    with pytest.raises(TypeError):
        # negative value is not ok in a unsigned int field
        ainst.c = -1

    # signed to unsigned is ok when value fits

    ainst.c = 42

    with pytest.raises(TypeError):
        # incompatible field type
        class B(HasTraits):
            a = Int(field_type=float)

    with pytest.raises(TypeError):
        # incompatible field default
        class B(HasTraits):
            a = Int(default=1.0)


def test_numerics_respect_choices_and_null():
    with pytest.raises(TraitValueError):
        # choices are respected by numeric attributes
        class B(HasTraits):
            a = Int(default=1, choices=(3, 4))

    class B(HasTraits):
        a = Int(default=1, choices=(1, 3, 4))
        odd_nullable_int = Int(required=False)

    with pytest.raises(TraitValueError):
        # choices are respected by numeric attributes
        binst = B()
        binst.a = 43

    binst = B()
    binst.odd_nullable_int = None

    with pytest.raises(TraitValueError):
        binst = B()
        binst.a = None


def test_float_attribute():
    class A(HasTraits):
        a = Float()
        b = Float(field_type=np.float32)
        c = Float(field_type=np.float16)

    ainst = A()
    # int's are ok
    ainst.a = 1
    ainst.a = int(2 ** 31 - 1)
    # larger floats as well if they actually fit
    ainst.c = np.float64(4)
    # they are converted to the declared types
    assert type(ainst.c) == np.float16

    with pytest.raises(TypeError):
        # out of bounds
        ainst.c = 2 ** 30


def test_deleting_a_declared_attribute_not_supported():
    class A(HasTraits):
        a = Attr(str)

    ainst = A()
    ainst.a = 'ana'

    with pytest.raises(AttributeError):
        del ainst.a


def test_dynamic_attributes_behave_statically_and_warn():
    class A(HasTraits):
        a = Attr(str)

    # these are logged warnings not errors. hard to test. here for coverage
    A.b = Attr(int)

    # the rest of the api assumes you don't change the attributes
    # this is the surprising result, no b in the declarative attr list
    assert set(A.declarative_attrs) == {'a', 'gid'}

    # this fails
    with pytest.raises(AttributeError):
        A().b

    class B(HasTraits):
        a = Attr(str)

    del B.a

    # the rest of the api assumes you don't change the attributes
    # this is the surprising result a not deleted from declarative attr list
    assert set(B.declarative_attrs) == {'a', 'gid'}


def test_declarative_properties():
    class A(HasTraits):
        def __init__(self):
            self._foo = None

        @trait_property(Attr(int))
        def xprop(self):
            return 23

        @trait_property(Attr(str))
        def foo(self):
            return self._foo

        @foo.setter
        def foo(self, val):
            self._foo = val

        @trait_property(Attr(int))
        def lyin_prop(self):
            return 'trickster'

    a = A()
    # read
    assert a.xprop == 23

    with pytest.raises(AttributeError):
        # nope read only
        a.xprop = 2

    with pytest.raises(AttributeError):
        # not supported
        del a.xprop

    # read-write

    a.foo = 'ana'
    assert a.foo == 'ana'

    # properties enforce traited types
    with pytest.raises(TypeError):
        a.foo = 42

    # on trying to read the lying prop we get an error
    with pytest.raises(TypeError):
        a.lyin_prop


def test_declarative_props_enforcing_shapes():
    class A(HasTraits):
        n_node = Int()

        def __init__(self, **kwargs):
            super(A, self).__init__(**kwargs)
            self._weights = None

        @trait_property(NArray(shape=(Dim.any, Dim.any)))
        def weights(self):
            return self._weights

        @weights.setter
        def weights(self, val):
            if val.shape != (self.n_node, self.n_node):
                raise TraitValueError
            self._weights = val

    a = A(n_node=4)
    a.weights = numpy.eye(4)

    with pytest.raises(TraitValueError):
        a.weights = numpy.zeros((2, 3))


def test_get_known_subclasses():
    class A(HasTraits):
        @abc.abstractmethod
        def frob(self):
            pass

    class B(A):
        def frob(self):
            pass

    class C(B):
        pass

    assert set(A.get_known_subclasses().values()) == {B, C}
    assert set(A.get_known_subclasses(include_abstract=True, include_itself=True).values()) == {A, B, C}
    assert set(A.get_known_subclasses(include_abstract=True).values()) == {B, C}
    assert set(A.get_known_subclasses(include_itself=True).values()) == {B, C}


def test_summary_info():
    class Z(HasTraits):
        zu = Attr(int)

    class A(HasTraits):
        a = Attr(str, default='ana')
        b = NArray(dtype=np.int64)
        ref = Attr(field_type=Z)

    ainst = A(b=np.arange(3))
    ainst.title = 'the red rose'
    ainst.gid = uuid.UUID(int=0)
    zinst = Z(zu=2)
    zinst.title = 'Z zuzu'
    ainst.ref = zinst
    summary = ainst.summary_info()

    assert summary['Type'] == 'A'
    assert summary['title'] == 'the red rose'
    assert summary['a'] == "'ana'"
    assert summary['b'] == ' [min, median, max] = [0, 1, 2] dtype = int64 shape = (3,)'
    assert summary['ref'] == 'Z zuzu'


def test_narray_summary_info():
    arr = np.array([1, 4, 9])
    summary = narray_summary_info(arr, ar_name='attribute_name')

    assert summary['attribute_name shape'] == '(3,)'
    assert summary['attribute_name dtype'][:3] == 'int'
    assert summary['attribute_name [min, median, max]'] == '[1, 4, 9]'


def test_narray_summary_info_none():
    summary = narray_summary_info(None, ar_name='attribute_name')
    assert summary['attribute_name is None'] == 'True'
    assert 1 == len(summary.keys())


def test_narray_summary_info_empty():
    summary = narray_summary_info(np.array([]), ar_name='attribute_name')
    assert summary['attribute_name is empty'] == 'True'
    assert 1 == len(summary.keys())


def test_narray_summary_info_without_arname():
    arr = np.array([1, 2, 3])
    summary = narray_summary_info(arr)

    assert summary['[min, median, max]'] == '[1, 2, 3]'
    assert summary['shape'] == "(3,)"


def test_narray_summary_info_condensed_form():
    arr = np.arange(3)
    summary = narray_summary_info(arr, ar_name='attribute_name', condensed=True)

    assert '[min, median, max] = [0, 1, 2]' in summary['attribute_name']
    assert 'dtype = int' in summary['attribute_name']
    assert 'shape = (3,)' in summary['attribute_name']


def test_narray_summary_info_condensed_form_single_item():
    arr = np.array([4])
    summary = narray_summary_info(arr, ar_name='attribute_name', condensed=True)

    assert summary['attribute_name'] == '4'


def test_narray_summary_info_with_nan():
    arr = np.array([1, 2, np.inf, np.nan])
    summary = narray_summary_info(arr)

    assert summary['has NaN'] == 'True'
    assert summary['shape'] == "(4,)"

    desc = narray_describe(arr)
    assert desc is not None


def test_hastraits_str_does_not_crash():
    class A(HasTraits):
        a = Attr(str, default='ana')
        b = NArray(dtype=int)
        pom = 'prun'

        @trait_property(Attr(int))
        def xprop(self):
            return 23

    ainst = A(b=np.arange(3))
    str(ainst)


def test_hastraits_html_repr_does_not_crash():
    class A(HasTraits):
        a = Attr(str, default='ana')
        b = NArray(dtype=int)

    ainst = A(b=np.arange(3))
    ainst._repr_html_()


def test_special_attributes_disallowed():
    with pytest.raises(TypeError):
        class A(HasTraits):
            _own_declarative_attrs = ('a')

    with pytest.raises(TypeError):
        class A(HasTraits):
            _own_declarative_props = ('a')


def test_linspacerange():
    ls = LinspaceRange(0, 10, 5)
    assert 1 in ls

    numpy.testing.assert_allclose(
        ls.to_array(),
        [0.0, 2.5, 5.0, 7.5, 10],
    )
    # test that repr will not crash
    repr(ls)


def test_range():
    ra = Range(0, 10, 5)
    assert 1 in ra

    numpy.testing.assert_allclose(
        ra.to_array(),
        [0, 5]
    )
    repr(ra)


def test_neotraits_ex():
    class A(HasTraits):
        foo = Int(default=1)

    try:
        a = A()
        a.foo = 'ana'
    except TraitTypeError as ex:
        assert ex.attr.field_name == 'foo'
        msg = str(ex)


def test_pedantic_edge_cases_for_coverage():
    class A(HasTraits):
        x = NArray()

    a = A()
    a.tag('subject', 'john doe')
    str(Attr(object))


def access_attr(a):
    a.s = 'has'
    a.arr = numpy.eye(3) + a.arr


@pytest.mark.benchmark
def test_perf_plain(benchmark):
    class Plain(object):
        def __init__(self):
            self.a = None
            self.arr = numpy.eye(3)

    benchmark(access_attr, Plain())


@pytest.mark.benchmark
def test_perf_trait(benchmark):
    class A(HasTraits):
        s = Attr(str, choices=('Ana', 'has', 'some', 'apples'))
        arr = NArray(shape=(Dim.any, Dim.any), default=numpy.eye(3))

    benchmark(access_attr, A())


def test_function_attribute():
    def func1():
        return 2.0

    def func2():
        return "Hello!"

    class A(HasTraits):
        a = Attr(field_type=types.FunctionType)
        b = Attr(field_type=types.FunctionType, default=lambda: 1)
        c = Attr(field_type=types.FunctionType, default=func1)

    ainst = A()
    ainst.a = func2
    val_a = ainst.a()
    val_b = ainst.b()
    val_c = ainst.c()
    assert val_a == "Hello!"
    assert val_b == 1
    assert val_c == 2.0

    with pytest.raises(TypeError):
        # out of bounds
        ainst.c = "Not a function"


def test_deepcopy():
    con = connectivity.Connectivity.from_file("connectivity_192.zip")
    original = simulator.Simulator(
        connectivity=con,
        coupling=coupling.Linear(a=numpy.array([2e-4])),
        integrator=integrators.EulerStochastic(dt=10.0),
        model=models.Linear(gamma=numpy.array([-1e-2])),
        monitors=(monitors.Raw(),),
        simulation_length=60e3
    )
    clone = original.duplicate()  # deepcopy() called inside duplicate()

    # random attrs
    o_random_stream = original.integrator.noise.random_stream
    c_random_stream = clone.integrator.noise.random_stream

    # check that attr are not the same
    assert original.integrator.noise != clone.integrator.noise
    assert original.integrator.noise != integrators.IntegratorStochastic.noise.default
    assert original.model != clone.model
    assert original.monitors[0] != clone.monitors[0]
    assert original.coupling != clone.coupling
    assert original.connectivity != clone.connectivity
    # even in the case of random generators
    assert o_random_stream is not c_random_stream

    # but values of attr are the same
    assert original.integrator.noise.ntau == clone.integrator.noise.ntau == 0
    assert original.integrator.noise.dt is clone.integrator.noise.dt is None
    assert original.model.gamma[0] == clone.model.gamma[0]
    assert original.coupling.a[0] == clone.coupling.a[0]
    assert original.connectivity.number_of_regions == clone.connectivity.number_of_regions == 0
    # random generators
    assert np.array_equal(o_random_stream.get_state()[1], c_random_stream.get_state()[1])
    assert o_random_stream.beta(1.0, 2.0) == c_random_stream.beta(1.0, 2.0)

    # configure simulation and modify original
    original.model.gamma[0] = 0.5
    original.integrator.noise.ntau = 2.0
    original.configure()

    # check that values of attr are now different and values from clone are did not change
    assert original.integrator.noise.ntau != clone.integrator.noise.ntau
    assert clone.integrator.noise.ntau == 0

    assert original.integrator.noise.dt != clone.integrator.noise.dt
    assert clone.integrator.noise.dt is None

    assert original.model.gamma[0] != clone.model.gamma[0]

    assert original.connectivity.number_of_regions != clone.connectivity.number_of_regions
    assert clone.connectivity.number_of_regions == 0

    # random generators
    assert not np.array_equal(o_random_stream.get_state()[1], c_random_stream.get_state()[1])
    assert o_random_stream.beta(1.0, 2.0) != c_random_stream.beta(1.0, 2.0)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#
"""
Created on Mar 20, 2013

.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

import numpy
from tvb.datatypes.spectral import WindowingFunctionsEnum
from tvb.tests.library.base_testcase import BaseTestCase
from tvb.datatypes import spectral, time_series


class TestSpectral(BaseTestCase):
    """
    Tests the defaults for `tvb.datatypes.spectral` module.
    """

    def test_fourierspectrum(self):
        data = numpy.random.random((10, 10))
        ts = time_series.TimeSeries(data=data, title='meh')
        dt = spectral.FourierSpectrum(source=ts,
                                      segment_length=100, array_data=numpy.array([]),
                                      windowing_function=WindowingFunctionsEnum.HAMMING)
        summary_info = dt.summary_info()
        assert summary_info['Frequency step'] == 0.01
        assert summary_info['Maximum frequency'] == 0.5
        assert summary_info['Segment length'] == 100
        assert summary_info['Windowing function'] is not None
        assert summary_info['Source'] == 'meh'
        assert summary_info['Spectral type'] == 'FourierSpectrum'
        assert dt.normalised_average_power is None
        assert dt.segment_length == 100.0
        assert dt.array_data is not None
        assert dt.source is not None
        assert dt.windowing_function is not None

    def test_waveletcoefficients(self):
        data = numpy.random.random((10, 10))
        ts = time_series.TimeSeries(data=data)
        dt = spectral.WaveletCoefficients(source=ts,
                                          mother='morlet',
                                          sample_period=7.8125,
                                          frequencies=numpy.array([0.008, 0.028, 0.048, 0.068]),
                                          normalisation="energy",
                                          q_ratio=5.0,
                                          array_data=numpy.random.random((10, 10)), )
        # dt.configure()
        summary_info = dt.summary_info()
        assert summary_info['Maximum frequency'] == 0.068
        assert summary_info['Minimum frequency'] == 0.008
        assert summary_info['Normalisation'], 'energy'
        assert summary_info['Number of scales'] == 4
        assert summary_info['Q-ratio'] == 5.0
        assert summary_info['Sample period'] == 7.8125
        assert summary_info['Spectral type'] == 'WaveletCoefficients'
        assert summary_info['Wavelet type'] == 'morlet'
        assert dt.q_ratio == 5.0
        assert dt.sample_period == 7.8125
        assert dt.array_data.shape == (10, 10)
        assert dt.source is not None

    def test_coherencespectrum(self):
        data = numpy.random.random((10, 10))
        ts = time_series.TimeSeries(data=data, title='meh')
        dt = spectral.CoherenceSpectrum(source=ts,
                                        nfft=4,
                                        array_data=numpy.random.random((10, 10)),
                                        frequency=numpy.random.random((10,)))
        summary_info = dt.summary_info()
        assert summary_info['Number of frequencies'] == 10
        assert summary_info['Spectral type'] == 'CoherenceSpectrum'
        assert summary_info['FFT length (time-points)'] == 4
        assert summary_info['Source'] == 'meh'
        assert dt.nfft == 4
        assert dt.array_data.shape == (10, 10)
        assert dt.source is not None

    def test_complexcoherence(self):
        data = numpy.random.random((10, 10))
        ts = time_series.TimeSeries(data=data, title='meh')
        dt = spectral.ComplexCoherenceSpectrum(source=ts,
                                               windowing_function=str(''),
                                               array_data=numpy.random.random((10, 10)),
                                               cross_spectrum=numpy.random.random((10, 10)),
                                               epoch_length=10,
                                               segment_length=5)
        summary_info = dt.summary_info()
        assert summary_info['Frequency step'] == 0.2
        assert summary_info['Maximum frequency'] == 0.5
        assert summary_info['Source'] == 'meh'
        assert summary_info['Spectral type'] == 'ComplexCoherenceSpectrum'
        assert dt.epoch_length == 10
        assert dt.segment_length == 5
        assert dt.array_data.shape, (10 == 10)
        assert dt.source is not None
        assert dt.windowing_function is not None



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
This file benchmarks isolated components in the scientific library.

"""

import sys
import time
import importlib
import numpy
import cpuinfo

print('CPU is a %s' % (cpuinfo.get_cpu_info()['brand'],))


def get_subclasses(submodule_name, base_class_name, ignore_names=None):
    modname = 'tvb.simulator.' + submodule_name
    mod = importlib.import_module(modname, modname)
    base_class = getattr(mod, base_class_name)
    ignore_names = [] if ignore_names is None else ignore_names
    for key in dir(mod):
        attr = getattr(mod, key)
        if (key != base_class_name
                and isinstance(attr, type)
                and issubclass(attr, base_class)
                and key not in ignore_names):
            yield attr


# integrators {{{

def integrators():
    return get_subclasses('integrators', 'Integrator', ignore_names=["IntegratorStochastic"])


def nop_dfun(x, coupling, local_coupling):
    return x


def eps_for_integrator(integrator_class, n_node, time_limit=0.5):
    integ = integrator_class()
    integ.configure()
    if 'Stochastic' in integrator_class.__name__:
        integ.noise.dt = integ.dt
    x = numpy.random.randn(n_node)
    thunk = lambda: integ.scheme(x, nop_dfun, None, None, 0.0)
    thunk()
    # start timing
    tic = time.time()
    n_eval = 0
    while (time.time() - tic) < time_limit:
        thunk()
        n_eval += 1
    toc = time.time()
    return n_eval / (toc - tic)


# }}}

# models {{{ 

def models():
    return get_subclasses('models', 'Model')


def randn_state_for_model(model, n_node):
    shape = (model.nvar, n_node, model.number_of_modes)
    state = numpy.random.randn(*shape)
    return state


def zero_coupling_for_model(model, n_node):
    n_cvar = len(model.cvar)
    shape = (n_cvar, n_node, model.number_of_modes)
    coupling = numpy.zeros(shape)
    return coupling


def eps_for_model(model_class, n_node, time_limit=0.5):
    model = model_class()
    model.configure()
    state = randn_state_for_model(model, n_node)
    coupling = zero_coupling_for_model(model, n_node)
    # throw one away in case of initialization
    model.dfun(state, coupling)
    # start timing
    tic = time.time()
    n_eval = 0
    while (time.time() - tic) < time_limit:
        model.dfun(state, coupling)
        n_eval += 1
    toc = time.time()
    return n_eval / (toc - tic)


# }}}

def eps_report_for_components(comps, eps_func):
    n_nodes = [2 << i for i in range(14)]
    sys.stdout.write('%30s' % ('n_node',))
    [sys.stdout.write('%06s' % (n,)) for n in n_nodes]
    sys.stdout.write('\n')
    sys.stdout.flush()
    for comp in comps:
        name = comp.__name__ if isinstance(comp, type) else comp.__class__.__name__
        sys.stdout.write('%30s' % (name,))
        for n_node in n_nodes:
            deps = eps_func(comp, n_node)
            sdeps = '%0.1f' % (deps / 1e3,)
            sys.stdout.write('%06s' % (sdeps,))
            sys.stdout.flush()
        sys.stdout.write('\n')
        sys.stdout.flush()


if __name__ == '__main__':
    print('units in kHz')
    print('benchmarking models')
    eps_report_for_components(models(), eps_for_model)
    print('benchmarking integrators')
    from tvb.simulator.integrators import RungeKutta4thOrderDeterministic

    integs = list(integrators()) + [RungeKutta4thOrderDeterministic]
    eps_report_for_components(integs, eps_for_integrator)

# vim: sw=4 sts=4 ai et foldmethod=marker



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
.. moduleauthor:: Dionysios Perdikis <dionperd@gmail.com>
.. moduleauthor:: Gabriel Florea <gabriel.florea@codemart.ro>
"""
import matplotlib
import numpy
from deprecated import deprecated
from matplotlib import pyplot, gridspec
from matplotlib.colors import Normalize
from six import string_types
from tvb.basic.logger.builder import get_logger
from tvb.datatypes.time_series import TimeSeries
from tvb.simulator.plot.base_plotter import BasePlotter
from tvb.simulator.plot.timeseries_interactive import TimeSeriesInteractivePlotter
from tvb.simulator.plot.power_spectra_interactive import PowerSpectraInteractive
from tvb.simulator.plot.utils import ensure_string, generate_region_labels, ensure_list, isequal_string, \
    time_spectral_analysis, raise_value_error

LOGGER = get_logger(__name__)


def assert_time(time, n_times, time_unit="ms", logger=None):
    if time_unit.find("ms"):
        dt = 0.001
    else:
        dt = 1.0
    try:
        time = numpy.array(time).flatten()
        n_time = len(time)
        if n_time > n_times:
            # self.logger.warning("Input time longer than data time points! Removing redundant tail time points!")
            time = time[:n_times]
        elif n_time < n_times:
            # self.logger.warning("Input time shorter than data time points! "
            #                     "Extending tail time points with the same average time step!")
            if n_time > 1:
                dt = numpy.mean(numpy.diff(time))
            n_extra_points = n_times - n_time
            start_time_point = time[-1] + dt
            end_time_point = start_time_point + n_extra_points * dt
            time = numpy.concatenate([time, numpy.arange(start_time_point, end_time_point, dt)])
    except:
        if logger:
            logger.warning("Setting a default time step vector manually! Input time: " + str(time))
        time = numpy.arange(0, n_times * dt, dt)
    return time


@deprecated(reason="Use tvb-widgets instead")
class TimeSeriesPlotter(BasePlotter):
    linestyle = "-"
    linewidth = 1
    marker = None
    markersize = 2
    markerfacecolor = None
    tick_font_size = 12
    print_ts_indices = True

    def __init__(self, config=None):
        super(TimeSeriesPlotter, self).__init__(config)
        self.interactive_plotter = None
        self.print_ts_indices = self.print_regions_indices
        self.HighlightingDataCursor = lambda *args, **kwargs: None
        if matplotlib.get_backend() in matplotlib.rcsetup.interactive_bk and self.config.MOUSE_HOOVER:
            try:
                from mpldatacursor import HighlightingDataCursor
                self.HighlightingDataCursor = HighlightingDataCursor
            except ImportError:
                self.config.MOUSE_HOOVER = False
                # self.logger.warning("Importing mpldatacursor failed! No highlighting functionality in plots!")
        else:
            # self.logger.warning("Noninteractive matplotlib backend! No highlighting functionality in plots!")
            self.config.MOUSE_HOOVER = False

    @property
    def line_format(self):
        return {"linestyle": self.linestyle, "linewidth": self.linewidth,
                "marker": self.marker, "markersize": self.markersize, "markerfacecolor": self.markerfacecolor}

    def _ts_plot(self, time, n_vars, nTS, n_times, time_unit, subplots, offset=0.0, data_lims=[]):

        time_unit = ensure_string(time_unit)
        data_fun = lambda data, time, icol: (data[icol], time, icol)

        def plot_ts(x, iTS, colors, labels):
            x, time, ivar = x
            time = assert_time(time, len(x[:, iTS]), time_unit, self.logger)
            try:
                return pyplot.plot(time, x[:, iTS], color=colors[iTS], label=labels[iTS], **self.line_format)
            except:
                self.logger.warning("Cannot convert labels' strings for line labels!")
                return pyplot.plot(time, x[:, iTS], color=colors[iTS], label=str(iTS), **self.line_format)

        def plot_ts_raster(x, iTS, colors, labels, offset):
            x, time, ivar = x
            time = assert_time(time, len(x[:, iTS]), time_unit, self.logger)
            try:
                return pyplot.plot(time, -x[:, iTS] + (offset * iTS + x[:, iTS].mean()), color=colors[iTS],
                                   label=labels[iTS], **self.line_format)
            except:
                self.logger.warning("Cannot convert labels' strings for line labels!")
                return pyplot.plot(time, -x[:, iTS] + offset * iTS, color=colors[iTS],
                                   label=str(iTS), **self.line_format)

        def axlabels_ts(labels, n_rows, irow, iTS):
            if irow == n_rows:
                pyplot.gca().set_xlabel("Time (" + time_unit + ")")
            if n_rows > 1:
                try:
                    pyplot.gca().set_ylabel(str(iTS) + "." + labels[iTS])
                except:
                    self.logger.warning("Cannot convert labels' strings for y axis labels!")
                    pyplot.gca().set_ylabel(str(iTS))

        def axlimits_ts(data_lims, time, icol):
            pyplot.gca().set_xlim([time[0], time[-1]])
            if n_rows > 1:
                pyplot.gca().set_ylim([data_lims[icol][0], data_lims[icol][1]])
            else:
                pyplot.autoscale(enable=True, axis='y', tight=True)

        def axYticks(labels, nTS, offsets=offset):
            pyplot.gca().set_yticks((offset * numpy.array([list(range(nTS))]).flatten()).tolist())
            try:
                pyplot.gca().set_yticklabels(labels.flatten().tolist())
            except:
                labels = generate_region_labels(nTS, [], "", True)
                self.logger.warning("Cannot convert region labels' strings for y axis ticks!")

        if offset > 0.0:
            plot_lines = lambda x, iTS, colors, labels: \
                plot_ts_raster(x, iTS, colors, labels, offset)
        else:
            plot_lines = lambda x, iTS, colors, labels: \
                plot_ts(x, iTS, colors, labels)
        this_axYticks = lambda labels, nTS: axYticks(labels, nTS, offset)
        if subplots:
            n_rows = nTS
            def_alpha = 1.0
        else:
            n_rows = 1
            def_alpha = 0.5
        subtitle_col = lambda subtitle: pyplot.gca().set_title(subtitle)
        subtitle = lambda iTS, labels: None
        projection = None
        axlabels = lambda labels, vars, n_vars, n_rows, irow, iTS: axlabels_ts(labels, n_rows, irow, iTS)
        axlimits = lambda data_lims, time, n_vars, icol: axlimits_ts(data_lims, time, icol)
        loopfun = lambda nTS, n_rows, icol: list(range(nTS))
        return data_fun, time, plot_lines, projection, n_rows, n_vars, def_alpha, loopfun, \
               subtitle, subtitle_col, axlabels, axlimits, this_axYticks

    def _trajectories_plot(self, n_dims, nTS, nSamples, subplots):
        data_fun = lambda data, time, icol: data

        def plot_traj_2D(x, iTS, colors, labels):
            x, y = x
            try:
                return pyplot.plot(x[:, iTS], y[:, iTS], color=colors[iTS], label=labels[iTS], **self.line_format)
            except:
                self.logger.warning("Cannot convert labels' strings for line labels!")
                return pyplot.plot(x[:, iTS], y[:, iTS], color=colors[iTS], label=str(iTS), **self.line_format)

        def plot_traj_3D(x, iTS, colors, labels):
            x, y, z = x
            try:
                return pyplot.plot(x[:, iTS], y[:, iTS], z[:, iTS], color=colors[iTS],
                                   label=labels[iTS], **self.line_format)
            except:
                self.logger.warning("Cannot convert labels' strings for line labels!")
                return pyplot.plot(x[:, iTS], y[:, iTS], z[:, iTS], color=colors[iTS],
                                   label=str(iTS), **self.line_format)

        def subtitle_traj(labels, iTS):
            try:
                if self.print_ts_indices:
                    pyplot.gca().set_title(str(iTS) + "." + labels[iTS])
                else:
                    pyplot.gca().set_title(labels[iTS])
            except:
                self.logger.warning("Cannot convert labels' strings for subplot titles!")
                pyplot.gca().set_title(str(iTS))

        def axlabels_traj(vars, n_vars):
            pyplot.gca().set_xlabel(vars[0])
            pyplot.gca().set_ylabel(vars[1])
            if n_vars > 2:
                pyplot.gca().set_zlabel(vars[2])

        def axlimits_traj(data_lims, n_vars):
            pyplot.gca().set_xlim([data_lims[0][0], data_lims[0][1]])
            pyplot.gca().set_ylim([data_lims[1][0], data_lims[1][1]])
            if n_vars > 2:
                pyplot.gca().set_zlim([data_lims[2][0], data_lims[2][1]])

        if n_dims == 2:
            plot_lines = lambda x, iTS, colors, labels: \
                plot_traj_2D(x, iTS, colors, labels)
            projection = None
        elif n_dims == 3:
            plot_lines = lambda x, iTS, colors, labels: \
                plot_traj_3D(x, iTS, colors, labels)
            projection = '3d'
        else:
            raise_value_error("Data dimensions are neigher 2D nor 3D!, but " + str(n_dims) + "D!", LOGGER)
        n_rows = 1
        n_cols = 1
        if subplots is None:
            # if nSamples > 1:
            n_rows = int(numpy.floor(numpy.sqrt(nTS)))
            n_cols = int(numpy.ceil((1.0 * nTS) / n_rows))
        elif isinstance(subplots, (list, tuple)):
            n_rows = subplots[0]
            n_cols = subplots[1]
            if n_rows * n_cols < nTS:
                raise_value_error("Not enough subplots for all time series:"
                                  "\nn_rows * n_cols = product(subplots) = product(" + str(subplots) + " = "
                                  + str(n_rows * n_cols) + "!", LOGGER)
        if n_rows * n_cols > 1:
            def_alpha = 0.5
            subtitle = lambda labels, iTS: subtitle_traj(labels, iTS)
            subtitle_col = lambda subtitles, icol: None
        else:
            def_alpha = 1.0
            subtitle = lambda labels, iTS: None
            subtitle_col = lambda subtitles, icol: pyplot.gca().set_title(pyplot.gcf().title)
        axlabels = lambda labels, vars, n_vars, n_rows, irow, iTS: axlabels_traj(vars, n_vars)
        axlimits = lambda data_lims, time, n_vars, icol: axlimits_traj(data_lims, n_vars)
        loopfun = lambda nTS, n_rows, icol: list(range(icol, nTS, n_rows))
        return data_fun, plot_lines, projection, n_rows, n_cols, def_alpha, loopfun, \
               subtitle, subtitle_col, axlabels, axlimits

    # TODO: refactor to not have the plot commands here
    def plot_ts(self, data, time=None, var_labels=[], mode="ts", subplots=None, special_idx=[],
                subtitles=[], labels=[], offset=0.5, time_unit="ms",
                title='Time series', figure_name=None, figsize=None):
        if not isinstance(figsize, (list, tuple)):
            figsize = self.config.LARGE_SIZE
        if isinstance(data, dict):
            var_labels = data.keys()
            data = data.values()
        elif isinstance(data, numpy.ndarray):
            if len(data.shape) < 3:
                if len(data.shape) < 2:
                    data = numpy.expand_dims(data, 1)
                data = numpy.expand_dims(data, 2)
                data = [data]
            else:
                # Assuming a structure of Time X Space X Variables X Samples
                data = [data[:, :, iv].squeeze() for iv in range(data.shape[2])]
        elif isinstance(data, (list, tuple)):
            data = ensure_list(data)
        else:
            raise_value_error("Input timeseries: %s \n" "is not on of one of the following types: "
                              "[numpy.ndarray, dict, list, tuple]" % str(data), LOGGER)
        n_vars = len(data)
        data_lims = []
        for id, d in enumerate(data):
            if isequal_string(mode, "raster"):
                data[id] = (d - d.mean(axis=0))
                drange = numpy.max(data[id].max(axis=0) - data[id].min(axis=0))
                data[id] = data[id] / drange  # zscore(d, axis=None)
            data_lims.append([d.min(), d.max()])
        data_shape = data[0].shape
        if len(data_shape) == 1:
            n_times = data_shape[0]
            nTS = 1
            for iV in range(n_vars):
                data[iV] = data[iV][:, numpy.newaxis]
        else:
            n_times, nTS = data_shape[:2]
        if len(data_shape) > 2:
            nSamples = data_shape[2]
        else:
            nSamples = 1
        if special_idx is None:
            special_idx = []
        n_special_idx = len(special_idx)
        if len(subtitles) == 0:
            subtitles = var_labels
        if isinstance(labels, list) and len(labels) == n_vars:
            labels = [generate_region_labels(nTS, label, ". ", self.print_ts_indices) for label in labels]
        else:
            labels = [generate_region_labels(nTS, labels, ". ", self.print_ts_indices) for _ in range(n_vars)]
        if isequal_string(mode, "traj"):
            data_fun, plot_lines, projection, n_rows, n_cols, def_alpha, loopfun, \
            subtitle, subtitle_col, axlabels, axlimits = \
                self._trajectories_plot(n_vars, nTS, nSamples, subplots)
        else:
            if isequal_string(mode, "raster"):
                data_fun, time, plot_lines, projection, n_rows, n_cols, def_alpha, loopfun, \
                subtitle, subtitle_col, axlabels, axlimits, axYticks = \
                    self._ts_plot(time, n_vars, nTS, n_times, time_unit, 0, offset, data_lims)

            else:
                data_fun, time, plot_lines, projection, n_rows, n_cols, def_alpha, loopfun, \
                subtitle, subtitle_col, axlabels, axlimits, axYticks = \
                    self._ts_plot(time, n_vars, nTS, n_times, time_unit, ensure_list(subplots)[0])
        alpha_ratio = 1.0 / nSamples
        alphas = numpy.maximum(numpy.array([def_alpha] * nTS) * alpha_ratio, 0.1)
        alphas[special_idx] = numpy.maximum(alpha_ratio, 0.1)
        if isequal_string(mode, "traj") and (n_cols * n_rows > 1):
            colors = numpy.zeros((nTS, 4))
            colors[special_idx] = \
                numpy.array([numpy.array([1.0, 0, 0, 1.0]) for _ in range(n_special_idx)]).reshape((n_special_idx, 4))
        else:
            cmap = matplotlib.cm.get_cmap('jet')
            colors = numpy.array([cmap(0.5 * iTS / nTS) for iTS in range(nTS)])
            colors[special_idx] = \
                numpy.array([cmap(1.0 - 0.25 * iTS / nTS) for iTS in range(n_special_idx)]).reshape((n_special_idx, 4))
        colors[:, 3] = alphas
        lines = []
        pyplot.figure(title, figsize=figsize)
        axes = []
        for icol in range(n_cols):
            if n_rows == 1:
                # If there are no more rows, create axis, and set its limits, labels and possible subtitle
                axes += ensure_list(pyplot.subplot(n_rows, n_cols, icol + 1, projection=projection))
                axlimits(data_lims, time, n_vars, icol)
                axlabels(labels[icol % n_vars], var_labels, n_vars, n_rows, 1, 0)
                pyplot.gca().set_title(subtitles[icol])
            for iTS in loopfun(nTS, n_rows, icol):
                if n_rows > 1:
                    # If there are more rows, create axes, and set their limits, labels and possible subtitles
                    axes += ensure_list(pyplot.subplot(n_rows, n_cols, iTS + 1, projection=projection))
                    axlimits(data_lims, time, n_vars, icol)
                    subtitle(labels[icol % n_vars], iTS)
                    axlabels(labels[icol % n_vars], var_labels, n_vars, n_rows, (iTS % n_rows) + 1, iTS)
                lines += ensure_list(plot_lines(data_fun(data, time, icol), iTS, colors, labels[icol % n_vars]))
            if isequal_string(mode, "raster"):  # set yticks as labels if this is a raster plot
                axYticks(labels[icol % n_vars], nTS)
                yticklabels = pyplot.gca().yaxis.get_ticklabels()
                self.tick_font_size = numpy.minimum(self.tick_font_size,
                                                    int(numpy.round(self.tick_font_size * 100.0 / nTS)))
                for iTS in range(nTS):
                    yticklabels[iTS].set_fontsize(self.tick_font_size)
                    if iTS in special_idx:
                        yticklabels[iTS].set_color(colors[iTS, :3].tolist() + [1])
                pyplot.gca().yaxis.set_ticklabels(yticklabels)
                pyplot.gca().invert_yaxis()

        if self.config.MOUSE_HOOVER:
            for line in lines:
                self.HighlightingDataCursor(line, formatter='{label}'.format, bbox=dict(fc='white'),
                                            arrowprops=dict(arrowstyle='simple', fc='white', alpha=0.5))

        pyplot.gcf().tight_layout()
        self._save_figure(pyplot.gcf(), figure_name)
        self._check_show()
        return pyplot.gcf(), axes, lines

    def plot_ts_raster(self, data, time, var_labels=[], time_unit="ms", special_idx=[],
                       title='Raster plot', subtitles=[], labels=[], offset=0.5, figure_name=None, figsize=None):
        if not isinstance(figsize, (list, tuple)):
            figsize = self.config.VERY_LARGE_SIZE
        return self.plot_ts(data, time, var_labels, "raster", None, special_idx, subtitles, labels, offset, time_unit,
                            title, figure_name, figsize)

    def plot_ts_trajectories(self, data, var_labels=[], subtitles=None, special_idx=[], labels=[],
                             title='State space trajectories', figure_name=None, figsize=None):
        if not isinstance(figsize, (list, tuple)):
            figsize = self.config.LARGE_SIZE
        return self.plot_ts(data, [], var_labels, "traj", subtitles, special_idx, labels=labels, title=title,
                            figure_name=figure_name, figsize=figsize)

    def plot_tvb_time_series(self, time_series, mode="ts", subplots=None, special_idx=[], subtitles=[],
                             offset=0.5, title=None, figure_name=None, figsize=None):
        if not isinstance(figsize, (list, tuple)):
            figsize = self.config.LARGE_SIZE
        if title is None:
            title = time_series.title
        variables_labels = time_series.labels_dimensions.get(time_series.labels_ordering[1], [])
        space_labels = time_series.labels_dimensions.get(time_series.labels_ordering[2], [])
        return self.plot_ts(numpy.swapaxes(time_series.data, 1, 2), time_series.time, variables_labels,
                            mode, subplots, special_idx, subtitles, space_labels,
                            offset, time_series.time_unit, title, figure_name, figsize)

    def plot_time_series(self, time_series, mode="ts", subplots=None, special_idx=[], subtitles=[],
                         offset=0.5, title=None, figure_name=None, figsize=None, **kwargs):
        if isinstance(time_series, TimeSeries):
            self.plot_tvb_time_series(time_series, mode, subplots, special_idx,
                                      subtitles, offset, title, figure_name, figsize)
        elif isinstance(time_series, (numpy.ndarray, dict, list, tuple)):
            time = kwargs.get("time", None)
            time_unit = kwargs.get("time_unit", "ms")
            labels = kwargs.get("labels", [])
            var_labels = kwargs.get("var_labels", [])
            if title is None:
                title = "Time Series"
            return self.plot_ts(time_series, time=time, mode=mode, time_unit=time_unit,
                                labels=labels, var_labels=var_labels, subplots=subplots, special_idx=special_idx,
                                subtitles=subtitles, offset=offset, title=title, figure_name=figure_name,
                                figsize=figsize)
        else:
            raise_value_error("Input time_series: %s \n" "is not on of one of the following types: [TimeSeries "
                              "(tvb-contrib), TimeSeries (tvb-library), numpy.ndarray, dict]" % str(time_series),
                              LOGGER)

    def plot_raster(self, time_series, subplots=None, special_idx=[], subtitles=[],
                    offset=0.5, title=None, figure_name=None, figsize=None, **kwargs):
        return self.plot_time_series(time_series, "raster", subplots, special_idx,
                                     subtitles, offset, title, figure_name, figsize, **kwargs)

    def plot_trajectories(self, time_series, subplots=None, special_idx=[], subtitles=[],
                          offset=0.5, title=None, figure_name=None, figsize=None, **kwargs):
        return self.plot_time_series(time_series, "traj", subplots, special_idx,
                                     subtitles, offset, title, figure_name, figsize, **kwargs)

    @staticmethod
    def plot_tvb_time_series_interactive(time_series, first_n=-1, **kwargs):
        interactive_plotter = TimeSeriesInteractivePlotter(time_series=time_series, first_n=first_n)
        interactive_plotter.configure()
        block = kwargs.pop("block", True)
        interactive_plotter.show(block=block, **kwargs)

    def plot_time_series_interactive(self, time_series, first_n=-1, **kwargs):
        if isinstance(time_series, TimeSeries):
            self.plot_tvb_time_series_interactive(time_series, first_n, **kwargs)
        elif isinstance(time_series, numpy.ndarray):
            self.plot_tvb_time_series_interactive(TimeSeries(data=time_series), first_n, **kwargs)
        elif isinstance(time_series, (list, tuple)):
            self.plot_tvb_time_series_interactive(TimeSeries(data=TimeSeries(data=numpy.stack(time_series, axis=1))),
                                                  first_n, **kwargs)
        elif isinstance(time_series, dict):
            ts = numpy.stack(time_series.values(), axis=1)
            time_series = TimeSeries(data=ts, labels_dimensions={"State Variable": time_series.keys()})
            self.plot_tvb_time_series_interactive(time_series, first_n, **kwargs)
        else:
            raise_value_error("Input time_series: %s \n" "is not on of one of the following types: [TimeSeries "
                              "(tvb-contrib), TimeSeriesTVB (tvb-library), numpy.ndarray, dict, list, tuple]" %
                              str(time_series), LOGGER)

    @staticmethod
    def plot_tvb_power_spectra_interactive(time_series, spectral_props, **kwargs):
        interactive_plotters = PowerSpectraInteractive(time_series=time_series, **spectral_props)
        interactive_plotters.configure()
        block = kwargs.pop("block", True)
        interactive_plotters.show(blocl=block, **kwargs)

    def plot_power_spectra_interactive(self, time_series, spectral_props, **kwargs):
        self.plot_tvb_power_spectra_interactive(self, time_series._tvb, spectral_props, **kwargs)

    # TODO: refactor to not have the plot commands here
    def _plot_ts_spectral_analysis_raster(self, data, time=None, var_label="", time_unit="ms",
                                          freq=None, spectral_options={}, special_idx=[], labels=[],
                                          title='Spectral Analysis', figure_name=None, figsize=None):
        if not isinstance(figsize, (list, tuple)):
            figsize = self.config.VERY_LARGE_SIZE
        if len(data.shape) == 1:
            n_times = data.shape[0]
            nS = 1
        else:
            n_times, nS = data.shape[:2]
        time = assert_time(time, n_times, time_unit, self.logger)
        if not isinstance(time_unit, string_types):
            time_unit = list(time_unit)[0]
        time_unit = ensure_string(time_unit)
        if time_unit in ("ms", "msec"):
            fs = 1000.0
        else:
            fs = 1.0
        fs = fs / numpy.mean(numpy.diff(time))
        n_special_idx = len(special_idx)
        if n_special_idx > 0:
            data = data[:, special_idx]
            nS = data.shape[1]
            if len(labels) > n_special_idx:
                labels = numpy.array([str(ilbl) + ". " + str(labels[ilbl]) for ilbl in special_idx])
            elif len(labels) == n_special_idx:
                labels = numpy.array([str(ilbl) + ". " + str(label) for ilbl, label in zip(special_idx, labels)])
            else:
                labels = numpy.array([str(ilbl) for ilbl in special_idx])
        else:
            if len(labels) != nS:
                labels = numpy.array([str(ilbl) for ilbl in range(nS)])
        if nS > 20:
            LOGGER.warning("It is not possible to plot spectral analysis plots for more than 20 signals!")
            return

        log_norm = spectral_options.get("log_norm", False)
        mode = spectral_options.get("mode", "psd")
        psd_label = mode
        if log_norm:
            psd_label = "log" + psd_label
        stf, time, freq, psd = time_spectral_analysis(data, fs,
                                                      freq=freq,
                                                      mode=mode,
                                                      nfft=spectral_options.get("nfft"),
                                                      window=spectral_options.get("window", 'hanning'),
                                                      nperseg=spectral_options.get("nperseg", int(numpy.round(fs / 4))),
                                                      detrend=spectral_options.get("detrend", 'constant'),
                                                      noverlap=spectral_options.get("noverlap"),
                                                      f_low=spectral_options.get("f_low", 10.0),
                                                      log_scale=spectral_options.get("log_scale", False))
        min_val = numpy.min(stf.flatten())
        max_val = numpy.max(stf.flatten())
        if nS > 2:
            figsize = self.config.VERY_LARGE_SIZE
        if len(var_label):
            title += ": " % var_label
        fig = pyplot.figure(title, figsize=figsize)
        fig.suptitle(title)
        gs = gridspec.GridSpec(nS, 23)
        ax = numpy.empty((nS, 2), dtype="O")
        img = numpy.empty((nS,), dtype="O")
        line = numpy.empty((nS,), dtype="O")
        for iS in range(nS, -1, -1):
            if iS < nS - 1:
                ax[iS, 0] = pyplot.subplot(gs[iS, :20], sharex=ax[iS, 0])
                ax[iS, 1] = pyplot.subplot(gs[iS, 20:22], sharex=ax[iS, 1], sharey=ax[iS, 0])
            else:
                # TODO: find and correct bug here
                ax[iS, 0] = pyplot.subplot(gs[iS, :20])
                ax[iS, 1] = pyplot.subplot(gs[iS, 20:22], sharey=ax[iS, 0])
            img[iS] = ax[iS, 0].imshow(numpy.squeeze(stf[:, :, iS]).T, cmap=pyplot.set_cmap('jet'),
                                       interpolation='none',
                                       norm=Normalize(vmin=min_val, vmax=max_val), aspect='auto', origin='lower',
                                       extent=(time.min(), time.max(), freq.min(), freq.max()))
            # img[iS].clim(min_val, max_val)
            ax[iS, 0].set_title(labels[iS])
            ax[iS, 0].set_ylabel("Frequency (Hz)")
            line[iS] = ax[iS, 1].plot(psd[:, iS], freq, 'k', label=labels[iS])
            pyplot.setp(ax[iS, 1].get_yticklabels(), visible=False)
            # ax[iS, 1].yaxis.tick_right()
            # ax[iS, 1].yaxis.set_ticks_position('both')
            if iS == (nS - 1):
                ax[iS, 0].set_xlabel("Time (" + time_unit + ")")

                ax[iS, 1].set_xlabel(psd_label)
            else:
                pyplot.setp(ax[iS, 0].get_xticklabels(), visible=False)
            pyplot.setp(ax[iS, 1].get_xticklabels(), visible=False)
            ax[iS, 0].autoscale(tight=True)
            ax[iS, 1].autoscale(tight=True)
        # make a color bar
        cax = pyplot.subplot(gs[:, 22])
        pyplot.colorbar(img[0], cax=pyplot.subplot(gs[:, 22]))  # fraction=0.046, pad=0.04) #fraction=0.15, shrink=1.0
        cax.set_title(psd_label)
        self._save_figure(pyplot.gcf(), figure_name)
        self._check_show()
        return fig, ax, img, line, time, freq, stf, psd

    def plot_ts_spectral_analysis_raster(self, data, time=None, time_unit="ms", freq=None, spectral_options={},
                                         special_idx=[], labels=[], title='Spectral Analysis', figure_name=None,
                                         figsize=None):
        if isinstance(data, dict):
            var_labels = data.keys()
            data = data.values()
        else:
            var_labels = []
            if isinstance(data, (list, tuple)):
                data = data[0]
            elif isinstance(data, numpy.ndarray) and data.ndim > 2:
                # Assuming a structure of Time X Space X Variables X Samples
                if data.ndim > 3:
                    data = data[:, :, :, 0]
                data = [data[:, :, iv].squeeze() for iv in range(data.shape[2])]
        if len(var_labels) == 0:
            var_labels = [""] * len(data)
        for d, var_label in zip(data, var_labels):
            self._plot_ts_spectral_analysis_raster(d, time, var_label, time_unit, freq, spectral_options,
                                                   special_idx, labels, title, figure_name, figsize)

    def plot_spectral_analysis_raster(self, time_series, freq=None, spectral_options={},
                                      special_idx=[], labels=[], title='Spectral Analysis', figure_name=None,
                                      figsize=None, **kwargs):
        if isinstance(time_series, TimeSeries):
            return self.plot_ts_spectral_analysis_raster(numpy.swapaxes(time_series.data, 1, 2).squeeze(),
                                                         time_series.time, time_series.time_unit, freq,
                                                         spectral_options,
                                                         special_idx, labels, title, figure_name, figsize)
        elif isinstance(time_series, (numpy.ndarray, dict, list, tuple)):
            time = kwargs.get("time", None)
            return self.plot_ts_spectral_analysis_raster(time_series, time=time, freq=freq,
                                                         spectral_options=spectral_options, special_idx=special_idx,
                                                         labels=labels, title=title, figure_name=figure_name,
                                                         figsize=figsize)
        else:
            raise_value_error("Input time_series: %s \n"
                              "is not on of one of the following types: [TimeSeries (tvb-contrib), "
                              "TimeSeries (tvb-library), numpy.ndarray, dict]" % str(time_series),
                              LOGGER)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

import os

from mako.template import Template
from mako.lookup import TemplateLookup
from mako.exceptions import text_error_template


here = os.path.dirname(os.path.abspath(__file__))


class MakoUtilMix:

    @property
    def lookup(self):
        lookup = TemplateLookup(directories=[here])
        return lookup

    def render_template(self, source, content):
        template = Template(source, lookup=self.lookup, strict_undefined=True)
        try:
            source = template.render(**content)
        except Exception as exc:
            print(text_error_template().render())
            raise exc
        return source

    def insert_line_numbers(self, source):
        lines = source.split('\n')
        numbers = range(1, len(lines) + 1)
        nu_lines = ['%03d\t%s' % (nu, li) for (nu, li) in zip(numbers, lines)]
        nu_source = '\n'.join(nu_lines)
        return nu_source



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
An interactive 3D head visualiser.

Usage
::

    #Create and launch the interactive visualiser
    from tvb.simulator.plot.head_plotter_3d import HeadPlotter3D
    hp = HeadPlotter3D()
    
    # To visualise the sensors and surface
    hp.display_source_sensor_geometry(surface, connectivity, meg_sensors, eeg_sensors)

    # To visualise a Surface in 3D
    hp.display_surface(surface)

"""

import matplotlib.pyplot as plt
from tvb.simulator.lab import *
from deprecated import deprecated
import ipywidgets as widgets
from IPython.display import display


@deprecated(reason="Use tvb-widgets instead")
class HeadPlotter3D(object):

    @staticmethod
    def display_source_sensor_geometry(surface=None, conn=None, meg_sensors=None, eeg_sensors=None):
        # type: (surfaces.SkinAir, connectivity.Connectivity, sensors.SensorsMEG, sensors.SensorsEEG ) -> None
        """
        :param surface: Optional surfaces.SkinAir instance. When none, we will try loading a default
        :param conn: Optional connectivity.Connectivity instance. When none, we will try loading a default
        :param meg_sensors: Optional sensors.SensorsMEG instance. When none, we will try loading a default
        :param eeg_sensors: Optional sensors.SensorsEEG instance. When none, we will try loading a default
        """

        if meg_sensors is None:
            meg_sensors = sensors.SensorsMEG.from_file()
        if eeg_sensors is None:
            eeg_sensors = sensors.SensorsEEG.from_file()
        if conn is None:
            conn = connectivity.Connectivity.from_file()
        if surface is None:
            surface = surfaces.SkinAir.from_file()

        # Configure Surface
        surface.configure()

        # Configure EEG Sensors
        eeg_sensors.configure()

        # WIDGET Controls and Layout
        box_layout = widgets.Layout(border='solid 1px black', margin='3px 3px 3px 3px', padding='2px 2px 2px 2px')
        params = dict()
        roi_checkbox = widgets.Checkbox(description="Show ROI Centers", value=False)
        eeg_checkbox = widgets.Checkbox(description="Show EEG Sensors", value=False)
        meg_checkbox = widgets.Checkbox(description="Show MEG Sensors", value=False)

        control_box = widgets.HBox([roi_checkbox, eeg_checkbox, meg_checkbox], layout=box_layout)

        # Connecting widgets with plot parameters
        params['ROI'] = roi_checkbox
        params['EEG'] = eeg_checkbox
        params['MEG'] = meg_checkbox

        fig = plt.figure()

        # Plotter Function
        def plot(**plot_params):
            fig.clf()
            ax = fig.add_subplot(111, projection='3d')

            # Plot boundary surface
            x, y, z = surface.vertices.T
            ax.plot_trisurf(x, y, z, triangles=surface.triangles, alpha=0.1, edgecolor='none')

            if plot_params['ROI']:
                # ROI centers as black circles
                x, y, z = conn.centres.T
                ax.plot(x, y, z, 'ko')

            if plot_params['EEG']:
                # EEG sensors as blue x's
                x, y, z = eeg_sensors.sensors_to_surface(surface).T
                ax.plot(x, y, z, 'bx')

            if plot_params['MEG']:
                # MEG sensors as red +'s
                x, y, z = meg_sensors.locations.T
                ax.plot(x, y, z, 'r+')

        out = widgets.interactive_output(plot, params)
        display(control_box, out)

    @staticmethod
    def display_surface(ctx):
        # type: (surfaces.CorticalSurface) -> None
        """
        :param ctx: surface.CorticalSurfaceF
        """
        ctx.configure()

        plt.figure()
        ax = plt.subplot(111, projection='3d')
        x, y, z = ctx.vertices.T
        ax.plot_trisurf(x, y, z, triangles=ctx.triangles, alpha=0.1, edgecolor='none')



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Tests for the NumPy backend.

.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""

import numpy as np

from tvb.simulator.backend.np import NpBackend
from tvb.simulator.coupling import Sigmoidal, Linear
from tvb.simulator.integrators import (EulerDeterministic, EulerStochastic,
    HeunDeterministic, HeunStochastic, IntegratorStochastic, 
    RungeKutta4thOrderDeterministic, Identity, IdentityStochastic,
    VODEStochastic)
from tvb.simulator.noise import Additive, Multiplicative
from tvb.datatypes.connectivity import Connectivity

from .backendtestbase import (BaseTestSim, BaseTestCoupling, BaseTestDfun,
    BaseTestIntegrate)


class TestNpSim(BaseTestSim):

    def _test_mpr(self, integrator, delays=False):
        sim = self._create_sim(
            integrator,
            inhom_mmpr=True,
            delays=delays,
            run_sim=False
        )

        (th, yh), = NpBackend().run_sim(sim, print_source=True)
        (t, y), = sim.run()

        self._check_match(y, yh)
        np.testing.assert_allclose(t, th)

    def _test_mvar(self, integrator):
        pass # TODO

    def _test_integrator(self, Integrator, delays=False, nsig=False):
        dt = 0.01
        if issubclass(Integrator, IntegratorStochastic):
            if nsig:
                integrator = Integrator(dt=dt, noise=Additive(nsig=np.r_[dt,dt*2]))
            else:
                integrator = Integrator(dt=dt, noise=Additive(nsig=np.r_[dt]))
            integrator.noise.dt = integrator.dt
        else:
            integrator = Integrator(dt=dt)
        if isinstance(integrator, (Identity, IdentityStochastic)):
            self._test_mvar(integrator, delays=delays)
        else:
            self._test_mpr(integrator, delays=delays)


    # TODO move to BaseTestSim to avoid duplicating all the methods
    def test_euler(self): self._test_integrator(EulerDeterministic)
    def test_eulers(self): self._test_integrator(EulerStochastic)
    def test_eulersn(self): self._test_integrator(EulerStochastic, nsig=True)
    def test_heun(self): self._test_integrator(HeunDeterministic)
    def test_heuns(self): self._test_integrator(HeunStochastic)
    def test_rk4(self): self._test_integrator(RungeKutta4thOrderDeterministic)
    # def test_id(self): self._test_integrator(Identity)
    # def test_ids(self): self._test_integrator(IdentityStochastic)

    def test_deuler(self): self._test_integrator(EulerDeterministic,
                                                 delays=True)

    def test_deulers(self): self._test_integrator(EulerStochastic,
                                                  delays=True)

    def test_dheun(self): self._test_integrator(HeunDeterministic,
                                                delays=True)

    def test_dheuns(self): self._test_integrator(HeunStochastic,
                                                 delays=True)

    def test_drk4(self): self._test_integrator(RungeKutta4thOrderDeterministic,
                                               delays=True)

    # def test_did(self): self._test_integrator(Identity,
    #                                           delays=True)
    #
    # def test_dids(self): self._test_integrator(IdentityStochastic,
    #                                            delays=True)

    def test_scipy_int_notimpl(self):
        with self.assertRaises(NotImplementedError):
            self._test_integrator(VODEStochastic)

    def test_multnoise_notimpl(self):
        dt = 0.01
        integrator = HeunStochastic(dt=dt, noise=Multiplicative(nsig=np.r_[dt]))
        with self.assertRaises(NotImplementedError):
            self._test_mpr(integrator)

class TestNpCoupling(BaseTestCoupling):

    def _test_cfun(self, cfun):
        "Test a Python cfun template."
        sim = self._prep_sim(cfun)
        # prep & invoke kernel
        template = f'''import numpy as np
<%include file="np-coupling.py.mako"/>
'''
        kernel = NpBackend().build_py_func(template, dict(sim=sim, np=np), 
            name='coupling', print_source=True)
        fill = np.r_[:sim.history.buffer.size]
        fill = np.reshape(fill, sim.history.buffer.shape[:-1])
        sim.history.buffer[..., 0] = fill
        sim.current_state[:] = fill[0,:,:,None]
        buf = sim.history.buffer[...,0]
        # kernel has history in reverse order except 1st element 🤕
        rbuf = np.concatenate((buf[0:1], buf[1:][::-1]), axis=0)
        state = np.transpose(rbuf, (1, 0, 2)).astype('f')
        weights = sim.connectivity.weights.astype('f')
        cX = np.zeros_like(state[:,0])
        kernel(cX, weights, state, sim.connectivity.delay_indices)
        # do comparison
        (t, y), = sim.run()
        np.testing.assert_allclose(cX, y[0,:,:,0], 1e-5, 1e-6)

    def test_np_linear(self): self._test_cfun(Linear())
    def test_np_sigmoidal(self): self._test_cfun(Sigmoidal())


class TestNpDfun(BaseTestDfun):

    def _test_dfun(self, model_):
        "Test a Python cfun template."
        class sim:  # dummy sim
            model = model_
        template = '''import numpy as np
<%include file="np-dfuns.py.mako"/>
'''
        kernel = NpBackend().build_py_func(template, dict(sim=sim, np=np),
            name='dfuns', print_source=True)
        cX = np.random.rand(2, 128)
        dX = np.zeros_like(cX)
        state = np.random.rand(2, 1, 128)
        parmat = sim.model.spatial_parameter_matrix
        kernel(dX, state, cX, parmat)
        np.testing.assert_allclose(dX,
                                   sim.model.dfun(state, cX)[:, 0])

    def test_py_mpr_symmetric(self):
        "Test symmetric MPR model"
        self._test_dfun(self._prep_model())

    def test_py_mpr_spatial1(self):
        "Test MPR w/ 1 spatial parameter."
        self._test_dfun(self._prep_model(1))

    def test_py_mpr_spatial2(self):
        "Test MPR w/ 2 spatial parameters."
        self._test_dfun(self._prep_model(2))


class TestNpIntegrate(BaseTestIntegrate):

    def _test_dfun(self, state, cX, lc):
        return -state*cX**2/state.shape[1]

    def _eval_cg(self, integrator_, state, weights_):
        class sim:
            integrator = integrator_
            connectivity = Connectivity.from_file()
            class model:
                state_variables = 'foo', 'bar'
        sim.connectivity.speed = np.r_[np.inf]
        sim.connectivity.configure()
        sim.integrator.configure()
        sim.connectivity.set_idelays(sim.integrator.dt)
        template = '''
import numpy as np
def coupling(cX, weights, state): cX[:] = weights.dot(state[:,0].T).T
def dfuns(dX, state, cX, parmat):
    d = -state*cX**2/state.shape[1]
    dX[:] = d
<%include file="np-integrate.py.mako" />
'''
        integrate = NpBackend().build_py_func(template, dict(sim=sim, np=np),
            name='integrate', print_source=True)
        parmat = np.zeros(0)
        dX = np.zeros((integrator_.n_dx,)+state[:,0].shape)
        cX = np.zeros_like(state[:,0])
        np.random.seed(42)
        args = state, weights_, parmat, dX, cX
        if isinstance(sim.integrator, IntegratorStochastic):
            args = args + (sim.integrator.noise.nsig, )
        integrate(*args)
        return state

    def _test_integrator(self, Integrator):
        if issubclass(Integrator, IntegratorStochastic):
            integrator = Integrator(dt=0.1, noise=Additive(nsig=np.r_[0.01]))
            integrator.noise.dt = integrator.dt
        else:
            integrator = Integrator(dt=0.1)
        nn = 76
        state = np.random.randn(2, 1, nn)
        weights = np.random.randn(nn, nn)
        cx = weights.dot(state[:,0].T).T
        assert cx.shape == (2, nn)
        expected = integrator.scheme(state[:,0], self._test_dfun, cx, 0, 0)
        actual = state.copy()
        np.random.seed(42)
        self._eval_cg(integrator, actual, weights)
        np.testing.assert_allclose(actual[:,0], expected)

    def test_euler(self): self._test_integrator(EulerDeterministic)
    def test_eulers(self): self._test_integrator(EulerStochastic)
    def test_heun(self): self._test_integrator(HeunDeterministic)
    def test_heuns(self): self._test_integrator(HeunStochastic)
    def test_rk4(self): self._test_integrator(RungeKutta4thOrderDeterministic)
    def test_id(self): self._test_integrator(Identity)
    def test_ids(self): self._test_integrator(IdentityStochastic)


# TODO monitor support

# TODO surface support

# TODO stimulus support

# TODO bounds/clamp support




# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Tests full region mapping when both surface and sub-cortical regions are present.

.. moduleauthor:: Borana Dollomaja <borana.dollomaja@univ-amu.fr>
"""

import numpy

from tvb.simulator.lab import *
from tvb.tests.library.base_testcase import BaseTestCase


class TestRegionMapping(BaseTestCase):

    def setUp(self, monitors=(monitors.Raw(),)):
        """
        Initialize the structural information, coupling function, integrator, 
        monitors, surface and stimulation.
        """
        # Connectome
        con = connectivity.Connectivity.from_file("connectivity_192.zip")
        con.configure()

        # Surface and local connectivity kernel
        surf = cortex.Cortex.from_file(local_connectivity_file="local_connectivity_16384.mat")
        surf.region_mapping_data.connectivity = con
        surf.configure()

        # Model
        oscilator = models.Generic2dOscillator()

        # monitors[0].spatial_mask = RegionMapping.full_region_mapping(surf, con)

        self.sim = simulator.Simulator(
            conduction_speed=1.0,
            coupling= coupling.Difference(a=numpy.array([0.01])),
            surface=surf,
            integrator=integrators.Identity(dt=1.0),
            simulation_length=10.0,
            connectivity=con,
            model=oscilator,
            monitors=monitors
        )
        self.sim.configure()

    def test_spatial_average_monitor(self):
        self.setUp(monitors=(monitors.SpatialAverage(period=1.),))
        result = self.sim.run(simulation_length = 10)
        assert result is not None


# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
A CUDA backend which uses templating to generate simulation
code, with PyCUDA as the driver.

.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""


try:
    import pycuda
    import pycuda.autoinit
    from pycuda.compiler import SourceModule
    import pycuda.driver as drv
    from pycuda.driver import Out, In, InOut
    pycuda_available = True
except Exception as exc:
	pycuda_available = False

from .templates import MakoUtilMix


class CuBackend(MakoUtilMix):

    def build_func(self, template_source, content, name='kernel', print_source=False):
        "Build and retrieve a Python function from template."
        source = self.render_template(template_source, content)
        if print_source:
            print(source)
        try:
            module = SourceModule(source)
        except pycuda.driver.CompileError as exc:
            print(self.insert_line_numbers(source))
            raise exc
        func = module.get_function(name)
        return func



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Calculate an FFT on a TimeSeries DataType and return a FourierSpectrum DataType.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

import numpy
import scipy.signal

from tvb.basic.logger.builder import get_logger
from tvb.basic.neotraits.info import narray_describe
from tvb.datatypes.spectral import FourierSpectrum

log = get_logger(__name__)


SUPPORTED_WINDOWING_FUNCTIONS = {
    'hamming': numpy.hamming,
    'bartlett': numpy.bartlett,
    'blackman': numpy.blackman,
    'hanning': numpy.hanning
}



"""
A module for calculating the FFT of a TimeSeries object of TVB and returning
a FourierSpectrum object. A segment length and windowing function can be
optionally specified. By default the time series is segmented into 1 second
blocks and no windowing function is applied.
"""


def compute_fast_fourier_transform(time_series, segment_length, window_function, detrend):
    """
    # type: (TimeSeries, float, function, bool) -> FourierSpectrum
    Calculate the FFT of time_series broken into segments of length
    segment_length and filtered by window_function.

    Parameters
    __________

    time_series : TimeSeries
    The TimeSeries to which the FFT is to be applied.

    segment_length : float
    The segment length determines the frequency resolution of the resulting power spectra -- longer
    windows produce finer frequency resolution

    window_function : WindowingFunctionsEnum
    Windowing functions can be applied before the FFT is performed. Default is None, possibilities are: 'hamming';
    'bartlett';'blackman'; and 'hanning'. See, numpy.<function_name>.

    detrend : bool
    Default is True, False means no detrending is performed on the time series.
    """

    tpts = time_series.data.shape[0]
    time_series_length = tpts * time_series.sample_period

    # Segment time-series, overlapping if necessary
    nseg = int(numpy.ceil(time_series_length / segment_length))
    if nseg > 1:
        seg_tpts = numpy.ceil(segment_length / time_series.sample_period)
        overlap = (seg_tpts * nseg - tpts) / (nseg - 1.0)
        starts = [max(seg * (seg_tpts - overlap), 0) for seg in range(nseg)]
        segments = [time_series.data[int(start):int(start) + int(seg_tpts)]
                    for start in starts]
        segments = [segment[:, :, :, :, numpy.newaxis] for segment in segments]
        ts = numpy.concatenate(segments, axis=4)
    else:
        segment_length = time_series_length
        ts = time_series.data[:, :, :, :, numpy.newaxis]
        seg_tpts = ts.shape[0]

    log.debug("Segment length being used is: %s" % segment_length)

    # Base-line correct the segmented time-series
    if detrend:
        ts = scipy.signal.detrend(ts, axis=0)
        log.debug("time_series " + narray_describe(ts))

    # Apply windowing function
    if window_function is not None:
        wf = SUPPORTED_WINDOWING_FUNCTIONS[window_function.value]
        window_mask = numpy.reshape(wf(int(seg_tpts)),
                                    (int(seg_tpts), 1, 1, 1, 1))
        ts = ts * window_mask

    # Calculate the FFT
    result = numpy.fft.fft(ts, axis=0)
    nfreq = result.shape[0] // 2
    result = result[1:nfreq + 1, :]

    log.debug("result " + narray_describe(result))

    spectra = FourierSpectrum(
        source=time_series,
        segment_length=segment_length,
        array_data=result,
        windowing_function=window_function
    )
    spectra.configure()

    return spectra



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
This private module implements the neotraits declarative machinery.
The basic Attribute Property and their automatic discovery by the Metaclass.
"""
import abc
import inspect
import typing
from tvb.basic.neotraits.ex import TraitTypeError, TraitAttributeError
from tvb.basic.neotraits.info import auto_docstring
from tvb.basic.logger.builder import get_logger

# a logger for the whole traits system
log = get_logger('tvb.traits')


class _Attr(object):
    """
    A private base class of Attributes
    Contains the minimum expected functionality by the declarative system.
    """
    def __init__(self):
        self.field_name = None  # type: typing.Optional[str]  # to be set by metaclass
        self.owner = None       # type: typing.Optional[MetaType]  # to be set by metaclass

    def _assert_have_field_name(self):
        """ check that the fields we expect the be set by metaclass have been set """
        if self.field_name is None:
            # this is the case if the descriptor is not in a class of type MetaType
            raise AttributeError("Declarative attributes can only be declared in subclasses of HasTraits")

    # subclass api

    def _post_bind_validate(self):
        # type: () -> None
        """
        Validates this instance of Attr.
        This is called just after field_name is set, by MetaType.
        We do checks here and not in init in order to give better error messages.
        Attr should be considered initialized only after this has run
        """

    # descriptor protocol

    def __get__(self, instance, owner):
        # type: (typing.Optional['HasTraits'], 'MetaType') -> typing.Any
        self._assert_have_field_name()
        if instance is None:
            # called from class, not an instance
            return self
        # data is stored on the instance in a field with the same name
        return instance.__dict__[self.field_name]


    def __set__(self, instance, value):
        # type: ('HasTraits', typing.Any) -> None
        self._assert_have_field_name()
        instance.__dict__[self.field_name] = value


    def __delete__(self, instance):
        raise TraitAttributeError("can't be deleted", attr=self)

    # A modest attempt of making Attr immutable

    def __setattr__(self, key, value):
        """ After owner is set disallow any field assignment """
        if getattr(self, 'owner', None) is not None:
            raise TraitAttributeError(
                "Can't change an Attr after it has been bound to a class."
                "Reusing Attr instances for different fields is not supported."
            )
        super(_Attr, self).__setattr__(key, value)


    def __delattr__(self, item):
        raise TraitAttributeError("Deleting an Attr field is not supported.")



class _Property(object):
    def __init__(self, fget, attr, fset=None):
        # type: (typing.Callable, _Attr, typing.Optional[typing.Callable]) -> None
        self.fget = fget
        self.fset = fset
        self.__doc__ = fget.__doc__
        self.attr = attr



class MetaType(abc.ABCMeta):
    """
    Metaclass for the declarative traits.
    We inherit ABCMeta so that the users may use @abstractmethod without having to
    deal with 2 meta-classes.
    Even though we do this we don't support the dynamic registration of subtypes to these abc's
    """

    # This is a python metaclass.
    # For an introduction see https://docs.python.org/2/reference/datamodel.html

    # here to avoid some hasattr; is None etc checks. And to make pycharm happy
    # should be harmless and shadowed by _declarative_attrs on the returned classes
    _own_declarative_attrs = ()  # type: typing.Tuple[str, ...] # name of all declarative fields on this class
    _own_declarative_props = ()  # type: typing.Tuple[str, ...]

    # A record of all the classes we have created.
    # note: As this holds references and not weakrefs it will prevent class garbage collection.
    #       Deleting classes would break get_known_subclasses and this cache
    __classes = {}  # type: typing.Dict[str, type]


    def get_known_subclasses(cls, include_abstract=False, include_itself=False):
        # type: (bool, bool) -> typing.Dict[str, typing.Type[MetaType]]
        """
        Returns all subclasses that exist *now*.
        New subclasses can be created after this call,
        after importing a new module or dynamically creating subclasses.
        Use with care. Use after most relevant modules have been imported.
        """
        ret = {}

        for k, c in cls.__classes.items():
            if issubclass(c, cls):
                if inspect.isabstract(c) and not include_abstract:
                    continue
                if c == cls and not include_itself:
                    continue
                ret.update({k: c})
        return ret

    def __walk_mro_inherit_declarations(cls, declaration):
        ret = []
        for super_cls in cls.mro():
            if isinstance(super_cls, MetaType):
                for attr_name in getattr(super_cls, declaration):
                    if attr_name not in ret:  # attr was overridden, don't duplicate
                        ret.append(attr_name)
        return tuple(ret)

    @property
    def declarative_attrs(cls):
        # type: () -> typing.Tuple[str, ...]
        """
        Gathers all the declared attributes, including the ones declared in superclasses.
        This is a meta-property common to all classes with this metatype
        """
        # We walk the mro here. This is in contrast with _own_declarative_attrs which is
        # not computed but cached by the metaclass on the class.
        # Caching is faster, but comes with the cost of taking care of the caches validity
        return cls.__walk_mro_inherit_declarations('_own_declarative_attrs')

    @property
    def declarative_props(cls):
        # type: () -> typing.Tuple[str, ...]
        """
        Gathers all the declared props, including the ones declared in superclasses.
        This is a meta-property common to all classes with this metatype
        """
        return cls.__walk_mro_inherit_declarations('_own_declarative_props')

    # here only to have a similar invocation like declarative_attrs
    # namely type(traited_instance).own_declarative_attrs
    # consider the traited_instance._own_declarative_attrs discouraged
    @property
    def own_declarative_attrs(cls):
        return cls._own_declarative_attrs


    def __new__(mcs, type_name, bases, namespace):
        """
        Gathers the names of all declarative fields.
        Tell each Attr of the name of the field it is bound to.
        """
        # gather all declarative attributes defined in the class to be constructed.
        # validate all declarations before constructing the new type
        attrs = []
        props = []

        for k, v in namespace.items():
            if isinstance(v, _Attr):
                attrs.append(k)
            elif isinstance(v, _Property):
                props.append(k)

        # record the names of the declarative attrs in the _own_declarative_attrs field
        if '_own_declarative_attrs' in namespace:
            raise TraitTypeError('class attribute _own_declarative_attrs is reserved in traited classes')
        if '_own_declarative_props' in namespace:
            raise TraitTypeError('class attribute _own_declarative_props is reserved in traited classes')

        namespace['_own_declarative_attrs'] = tuple(attrs)
        namespace['_own_declarative_props'] = tuple(props)
        # construct the class
        cls = super(MetaType, mcs).__new__(mcs, type_name, bases, namespace)

        # inform the Attr instances about the class their are bound to
        for attr_name in attrs:
            v = namespace[attr_name]
            v.field_name = attr_name
            v.owner = cls
            # noinspection PyProtectedMember
            v._post_bind_validate()

        # do the same for props.
        for prop_name in props:
            v = namespace[prop_name].attr
            v.field_name = prop_name
            v.owner = cls
            # noinspection PyProtectedMember
            v._post_bind_validate()

        # update docstring. Note that this is only possible if cls was created by a metatype in python
        setattr(cls, '__doc_old__', cls.__doc__)
        setattr(cls, '__doc__', auto_docstring(cls))

        # update the HasTraits class registry
        mcs.__classes.update({cls.__name__: cls})
        return cls


    # note: Any methods defined here are metamethods, visible from all classes with this metatype
    # AClass.field if not found on AClass will be looked up on the metaclass
    # If you just want a regular private method here name it def __lala(cls)
    # double __ not for any privacy but to reduce namespace pollution
    # double __ will mangle the names and make such lookups fail

    # warn about dynamic Attributes


    def __setattr__(self, key, value):
        """
        Complain if TraitedClass.a = Attr()
        Traits assumes that all attributes are statically declared in the class body
        """
        if isinstance(value, _Attr):
            log.warning('dynamically assigned Attributes are not supported')
        super(MetaType, self).__setattr__(key, value)


    def __delattr__(self, item):
        if isinstance(getattr(self, item, None), _Attr):
            log.warning('Dynamically removing Attributes is not supported')
        super(MetaType, self).__delattr__(item)



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
Simulator history implementations.

.. moduleauthor:: Mihai Andrei <mihai.andrei@codemart.ro>
.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""


import numpy
from tvb.simulator.common import get_logger
from .descriptors import StaticAttr, Dim, NDArray
from .backend.ref import ReferenceBackend

LOG = get_logger(__name__)


class BaseHistory(StaticAttr):
    "Abstract base class for history implementations."

    n_time, n_node, n_cvar, n_mode = Dim(), Dim(), Dim(), Dim()

    weights = NDArray((n_node, n_node), 'f') # type: numpy.ndarray
    delays = NDArray((n_node, n_node), 'f') # type: numpy.ndarray
    cvars = NDArray((n_cvar, ), 'i') # type: numpy.ndarray

    @property
    def nbytes(self):
        arrays = 'weights delays cvars'.split()
        return sum([getattr(self, ary).nbytes for ary in arrays])

    def __init__(self, weights, delays, cvars, n_mode):
        self.n_time, self.n_cvar, self.n_node, self.n_mode = delays.max() + 1, len(cvars), delays.shape[0], n_mode
        self.weights = weights
        self.delays = delays
        self.cvars = cvars

    def initialize(self, init):
        raise NotImplemented

    def update(self, step, new_state):
        raise NotImplemented

    def query(self, step, out=None):
        raise NotImplemented

    @classmethod
    def from_simulator(cls, sim, initial_conditions=None):
        """
        Set initial conditions for the simulation using either the provided
        initial_conditions or, if none are provided, the model's initial()
        method. This method is called durin the Simulator's __init__().

        Any initial_conditions that are provided as an argument are expected
        to have dimensions 1, 2, and 3 with shapse corresponding to the number
        of state_variables, nodes and modes, respectively. If the provided
        inital_conditions are shorter in time (dim=0) than the required history
        the model's initial() method is called to make up the difference.

        """

        backend = ReferenceBackend
        initial_conditions = initial_conditions or sim.initial_conditions
        if initial_conditions is None:
            n_time, n_svar, n_node, n_mode = sim.good_history_shape
            sim.log.info('Preparing initial history of shape %r using model.initial()', sim.good_history_shape)
            if sim.surface is not None:
                n_node = sim.number_of_nodes
            history = sim.model.initial_for_simulator(sim.integrator, (n_time, n_svar, n_node, n_mode))
            initial_conditions = history[n_time - 1]
        # ICs provided
        else:
            # history should be [timepoints, state_variables, nodes, modes]
            sim.log.info('Using provided initial history of shape %r', initial_conditions.shape)
            n_time, n_svar, n_node, n_mode = ic_shape = initial_conditions.shape
            nr = sim.connectivity.number_of_regions
            if sim.surface is not None and n_node == nr:
                initial_conditions = initial_conditions[:, :, sim.surface.region_mapping]
                return sim._configure_history(initial_conditions)
            elif sim.surface is None and ic_shape[1:] != sim.good_history_shape[1:]:
                raise ValueError("Incorrect history sample shape %s, expected %s"
                                 % (ic_shape[1:], sim.good_history_shape[1:]))
            else:
                if ic_shape[0] >= sim.connectivity.horizon:
                    sim.log.debug("Using last %d time-steps for history.", sim.connectivity.horizon)
                    history = initial_conditions[-sim.connectivity.horizon:, :, :, :].copy()
                else:
                    sim.log.debug('Padding initial conditions with model.initial')
                    history = sim.model.initial_for_simulator(sim.integrator, sim.good_history_shape)
                    shift = sim.current_step % sim.connectivity.horizon
                    history = numpy.roll(history, -shift, axis=0)
                    if sim.surface is not None:
                        n_reg = sim.connectivity.number_of_regions
                        (nt, ns, _, nm), ax = history.shape, (2, 0, 1, 3)
                        region_initial_conditions = numpy.zeros((nt, ns, n_reg, nm))
                        backend.add_at(region_initial_conditions.transpose(ax), sim.surface.region_mapping, initial_conditions.transpose(ax))
                        region_initial_conditions /= numpy.bincount(sim.surface.region_mapping).reshape((-1, 1))
                        history[:region_initial_conditions.shape[0], :, :, :] = region_initial_conditions
                    else:
                        history[:ic_shape[0], :, :, :] = initial_conditions
                    history = numpy.roll(history, shift, axis=0)
                sim.current_step += ic_shape[0] - 1

        # Make sure that history values are bounded
        for it in range(history.shape[0]):
            sim.integrator.bound_and_clamp(history[it])
        sim.log.info('Final initial history shape is %r', history.shape)

        # create initial state
        if sim.surface:
            # ensure 4D
            # TODO refactor to backend
            initial_conditions = initial_conditions.reshape((-1, ) + initial_conditions.shape[-3:])
            sim.current_state = initial_conditions[-1].copy()
        else:
            sim.current_state = history[sim.current_step % sim.connectivity.horizon].copy()
        sim.log.info('initial state has shape %r' % (sim.current_state.shape, ))

        # create history buffer
        if sim.surface is not None and history.shape[2] > sim.connectivity.number_of_regions:
            n_reg = sim.connectivity.number_of_regions
            (nt, ns, _, nm), ax = history.shape, (2, 0, 1, 3)
            region_history = numpy.zeros((nt, ns, n_reg, nm))
            backend.add_at(region_history.transpose(ax), sim.surface.region_mapping, history.transpose(ax))
            region_history /= numpy.bincount(sim.surface.region_mapping).reshape((-1, 1))
            history = region_history

        # init history instance
        inst = cls(sim.connectivity.weights, sim.connectivity.idelays,
                   sim.model.cvar, sim.model.number_of_modes)
        inst.initialize(history)
        return inst


class DenseHistory(BaseHistory):
    "TVB's traditional history implementation."

    # extended shape arrays for indexing
    _es = 'n_node', 'n_cvar', 'n_node'
    es_icvar = NDArray(_es, 'i')
    es_idelays = NDArray(_es, 'i')
    es_weights = NDArray(_es + ('n_mode', ), 'f')
    es_node_ids = NDArray(_es, 'i')
    buffer = NDArray(('n_time', 'n_cvar', 'n_node', 'n_mode'), 'f', read_only=False)
    current_state = NDArray(('n_cvar', 'n_node', 'n_mode'), 'f', read_only=False)
    delayed_state = NDArray(('n_node', 'n_cvar', 'n_node', 'n_mode'), 'f', read_only=False)

    @property
    def nbytes(self):
        arrays = 'icvar idelays weights node_ids'.split()
        nbytes = sum([getattr(self, 'es_' + ary).nbytes for ary in arrays])
        nbytes += self.buffer.nbytes
        nbytes += BaseHistory.nbytes.fget(self)
        return nbytes

    def __init__(self, weights, delays, cvars, n_mode):
        super(DenseHistory, self).__init__(weights, delays, cvars, n_mode)

        # initialize indexing arrays
        na = numpy.newaxis
        self.es_icvar = numpy.r_[:len(self.cvars)][na, :, na]
        self.es_idelays = self.delays[:, na, :].astype('i')
        self.es_weights = self.weights[:, na, :, na]
        self.es_node_ids = numpy.r_[:self.n_node][na, na, :]

    def initialize(self, init):
        if init.shape[1] > len(self.cvars):
            init = init[:, self.cvars] # simulator still thinks history is (time, svar, ..)
        self.buffer = init

    def query(self, step, out=None):
        time_idx = (step - 1 - self.es_idelays + self.n_time) % self.n_time
        self.delayed_state = self.buffer[time_idx, self.es_icvar, self.es_node_ids]
        self.current_state = self.buffer[(step - 1) % self.n_time]
        return self.current_state, self.delayed_state

    def update(self, step, new_state):
        self.buffer[step % self.n_time] = new_state[self.cvars]


class SparseHistory(DenseHistory):
    "History implementation which stores data only for non-zero weights."

    n_nnzw = Dim()
    n_nnzr = Dim()
    time_stride = Dim()
    nnz_mask = NDArray(('n_node', 'n_node'), numpy.bool_)
    const_indices = NDArray(('n_cvar', n_nnzw, 'n_mode'), 'i')
    nnz_idelays = NDArray((n_nnzw,), 'i')
    nnz_row_el_idx = NDArray((n_nnzw, ), 'i')
    nnz_col_el_idx = NDArray((n_nnzw, ), 'i')
    nnz_weights = NDArray((n_nnzw, ), 'f')
    nnz_row_idx = NDArray((n_nnzr, ), 'i')

    def __init__(self, weights, delays, cvars, n_mode):
        super(SparseHistory, self).__init__(weights, delays, cvars, n_mode)
        self.time_stride = self.n_cvar * self.n_node * self.n_mode
        self.nnz_mask = weights_nonzero = weights != 0.0 # type: numpy.ndarray
        self.n_nnzw = nnz = weights_nonzero.sum()
        self.nnz_weights = weights[self.nnz_mask]
        self.nnz_row_el_idx, self.nnz_col_el_idx = numpy.argwhere(self.nnz_mask).T
        nnz_row_idx = numpy.unique(self.nnz_row_el_idx)
        self.n_nnzr = len(nnz_row_idx)
        self.nnz_row_idx = nnz_row_idx
        self.nnz_idelays = delays[weights_nonzero].astype('i')
        # build const indices
        n, m = self.n_node, self.n_mode
        icvars_ = numpy.r_[:len(cvars)].reshape((-1, 1, 1)) * n * m
        nodes_ = numpy.tile(numpy.r_[:n], (n, 1))[self.nnz_mask, numpy.newaxis] * m
        modes_ = numpy.r_[:m]
        self.const_indices = icvars_ + nodes_ + modes_
        self.delayed_state[:] = 0.0

        LOG.info('history has n_time=%d n_cvar=%d n_node=%d n_nmode=%d, requires %.2f MB',
                 self.n_time, self.n_cvar, self.n_node, self.n_mode, self.nbytes*2**-20)
        LOG.debug('sparse flat time_stride=%d', self.time_stride)
        LOG.info('sparse history has n_nnzw=%d, i.e. %.2f %% sparse', self.n_nnzw,
                 self.n_nnzw * 100.0 / self.n_node**2)

    def query(self, step, out=None):
        current, delayed = self.query_sparse(step)
        self.delayed_state.transpose((1, 0, 2, 3))[:, self.nnz_mask] = delayed
        return current, self.delayed_state

    def query_sparse(self, step):
        time_indices = ((step - 1 - self.nnz_idelays + self.n_time) % self.n_time) # type: numpy.ndarray
        time_indices = time_indices.reshape((-1, 1)) * self.time_stride # type: numpy.ndarray
        delayed_state = self.buffer.take(time_indices + self.const_indices)
        current_state = self.buffer[(step - 1) % self.n_time]
        return current_state, delayed_state

    @property
    def nbytes(self):
        arrays = 'nnz_mask const_indices nnz_idelays nnz_row_el_idx nnz_col_el_idx nnz_weights nnz_row_idx'.split()
        nbytes = sum([getattr(self, ary).nbytes for ary in arrays])
        nbytes += DenseHistory.nbytes.fget(self)
        return nbytes


# implement in order  NumPy, Numba & OpenCL versions

# simulator.history becomes impl instance

# state must also transpose for performance reasons

# bench history impl like other components

# trace history accesses

# cfun must also now expect to operate on (nnz, ncvar, nmode)


# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
A plain NumPy backend which uses templating to generate simulation
code.

.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""

import os
import sys
import importlib
import numpy as np
import autopep8
import tempfile

from .templates import MakoUtilMix

from tvb.simulator.lab import *


class NpBackend(MakoUtilMix):

    def __init__(self):
        self.cgdir = tempfile.TemporaryDirectory()
        sys.path.append(self.cgdir.name)

    def build_py_func(self, template_source, content, name='kernel', print_source=False,
            modname=None, fname=None):
        "Build and retrieve one or more Python functions from template."
        source = self.render_template(template_source, content)
        source = autopep8.fix_code(source)
        if print_source:
            print(self.insert_line_numbers(source))
        if fname is not None:
            fullfname = os.path.join(self.cgdir.name, fname)
            with open(fullfname, 'w') as fd:
                fd.write(source)
        if modname is not None:
            return self.eval_module(source, name, modname)
        else:
            return self.eval_source(source, name, print_source)

    def eval_source(self, source, name, print_source):
        globals_ = {}
        try:
            exec(source, globals_)
        except Exception as exc:
            if not print_source:
                print(self._insert_line_numbers(source))
            raise exc
        fns = [globals_[n] for n in name.split(',')]
        return fns[0] if len(fns)==1 else fns

    def eval_module(self, source, name, modname):
        here = os.path.abspath(os.path.dirname(__file__))
        genp = os.path.join(here, 'templates', 'generated')
        with open(f'{genp}/{modname}.py', 'w') as fd:
            fd.write(source)
        fullmodname = f'tvb.simulator.backend.templates.generated.{modname}'
        mod = importlib.import_module(fullmodname)
        fns = [getattr(mod,n) for n in name.split(',')]
        return fns[0] if len(fns)==1 else fns

    def _check_choices( self, val, choices):
        if not isinstance(val, choices):
            raise NotImplementedError("Unsupported simulator component. Given: {}\nExpected one of: {}".format(val, choices))

    def check_compatibility(self,sim):
        # monitors
        if len(sim.monitors) > 1:
            raise NotImplementedError("Configure with one monitor.")
        self._check_choices(sim.monitors[0], monitors.Raw)
        # integrators
        self._check_choices(sim.integrator, 
                (
                    integrators.HeunStochastic,
                    integrators.HeunDeterministic,
                    integrators.EulerStochastic,
                    integrators.EulerDeterministic,
                    integrators.Identity,
                    integrators.IdentityStochastic,
                    integrators.RungeKutta4thOrderDeterministic,
                )
        )
        # models 
        if sim.model.number_of_modes > 1:
            # this is a limitation e.g. by how nsig is now handled
            raise NotImplementedError("Only models with 1 mode are supported")
        self._check_choices(sim.model, models.MontbrioPazoRoxin) 
        # coupling
        self._check_choices(sim.coupling, 
                (coupling.Linear, coupling.Sigmoidal))
        # surface
        if sim.surface is not None:
            raise NotImplementedError("Surface simulation not supported.")
        # stimulus evaluated outside the backend, no restrictions

    def run_sim(self, sim, nstep=None, simulation_length=None, print_source=False):
        assert nstep is not None or simulation_length is not None or sim.simulation_length is not None

        self.check_compatibility(sim)

        if nstep is None:
            if simulation_length is None:
                simulation_length = sim.simulation_length
            nstep = int(np.ceil(simulation_length/sim.integrator.dt))

        buf = sim.history.buffer[...,0]
        rbuf = np.concatenate((buf[0:1], buf[1:][::-1]), axis=0)
        state = np.transpose(rbuf, (1, 0, 2)).astype('f')
        t = np.arange(1, nstep+1 ) * sim.integrator.dt 

        template = '<%include file="np-sim.py.mako"/>'
        content = dict(sim=sim, np=np, nstep=nstep)
        kernel = self.build_py_func(template, content, print_source=print_source)
        dX = state.copy()
        n_svar, _, n_node = state.shape
        state = state.reshape((n_svar, sim.connectivity.horizon, n_node))

        weights = sim.connectivity.weights.copy()
        yh = np.empty((len(t),)+state[:,0].shape)

        parmat = sim.model.spatial_parameter_matrix
        args = state, weights, yh, parmat
        if isinstance(sim.integrator, integrators.IntegratorStochastic):
            np.random.seed(sim.integrator.noise.noise_seed) 
            if len(sim.integrator.noise.nsig.shape) > 1:
                nsig = sim.integrator.noise.nsig[:,0] # no modes for now
            else:
                nsig = sim.integrator.noise.nsig
            args = args + (nsig,)
        if sim.connectivity.has_delays:
            args = args + (sim.connectivity.delay_indices,)
        kernel(*args)

        return (t, yh), 



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""

import numpy
from tvb.datatypes.sensors import SensorsEEG, SensorsMEG
from tvb.tests.library.base_testcase import BaseTestCase
from tvb.datatypes import time_series


class TestTimeseries(BaseTestCase):
    """
    Tests the defaults for `tvb.datatypes.time_series` module.
    """

    def test_timeseries(self):
        data = numpy.random.random((10, 10))
        dt = time_series.TimeSeries(data=data, title='meh if it counts')
        summary_info = dt.summary_info()
        assert summary_info['Dimensions'] == ('Time', 'State Variable', 'Space', 'Mode')
        assert summary_info['Length'] == 10.0
        assert summary_info['Sample period'] == 1.0
        assert summary_info['Time units'] == 'ms'
        assert summary_info['Time-series name'] == 'meh if it counts'
        assert summary_info['Time-series type'] == 'TimeSeries'
        assert dt.data.shape == (10, 10)
        assert dt.sample_period == 1.0
        assert dt.sample_rate == 1000
        assert dt.start_time == 0.0
        assert dt.time is None

    def test_timeserieseeg(self):
        data = numpy.random.random((10, 10))
        dt = time_series.TimeSeriesEEG(data=data, sensors=SensorsEEG())
        assert dt.data.shape == (10, 10)
        assert ("Time", "SV", "EEG Sensor", "Mode") == dt.labels_ordering
        assert dt.sample_period == 1.0
        assert dt.sample_rate == 1000
        assert dt.sensors is not None
        assert dt.start_time == 0.0
        assert dt.time is None

    def test_timeseriesmeg(self):
        data = numpy.random.random((10, 10))
        dt = time_series.TimeSeriesMEG(data=data, sensors=SensorsMEG(orientations=numpy.array([])))
        assert dt.data.shape == (10, 10)
        assert ("Time", "SV", "MEG Sensor", "Mode") == dt.labels_ordering
        assert dt.sample_period == 1.0
        assert dt.sample_rate == 1000
        assert dt.sensors is not None
        assert dt.start_time == 0.0
        assert dt.time is None

    def test_timeseriesregion(self):
        data = numpy.random.random((10, 10))
        dt = time_series.TimeSeriesRegion(data=data)
        assert dt.data.shape == (10, 10)
        assert dt.labels_ordering == ('Time', 'State Variable', 'Region', 'Mode')
        assert dt.sample_period == 1.0
        assert dt.sample_rate == 1000
        assert dt.start_time == 0.0
        assert dt.time is None

    def test_timeseriessurface(self):
        data = numpy.random.random((10, 10))
        dt = time_series.TimeSeriesSurface(data=data)
        assert dt.data.shape == (10, 10)
        assert dt.labels_ordering == ('Time', 'State Variable', 'Vertex', 'Mode')
        assert dt.sample_period == 1.0
        assert dt.sample_rate == 1000
        assert dt.start_time == 0.0
        assert dt.time is None

    def test_timeseriesvolume(self):
        data = numpy.random.random((10, 10))
        dt = time_series.TimeSeriesVolume(data=data)
        assert dt.data.shape == (10, 10)
        assert dt.labels_ordering == ('Time', 'X', 'Y', 'Z')
        assert dt.sample_period == 1.0
        assert dt.sample_rate == 1000
        assert dt.start_time == 0.0
        assert dt.time is None



# -*- coding: utf-8 -*-
#
#
# TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#


"""
This is the root of TVB tests. It overlaps between tvb-library and tvb-framework packages.

For running test use the command: pytest [tests_folder]

IMPORTANT: To ensure the correct TVB Profile is set in tests, before ANY tvb import in the unit test,
    setup the correct tvb test profile::

        from tvb.tests.library import setup_test_console_env
        setup_test_console_env()

        # OR
        from tvb.tests.framework.core.base_testcase import init_test_env
        init_test_env()

    You can do this implicitly (as done currently in the majority of our example unit tests,
    by importing BaseTestCase FIRST::

        from tvb.tests.library.base_testcase import BaseTestCase
        # OR
        from tvb.tests.framework.core.base_testcase import BaseTestCase
"""

from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2024, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as explained here:
# https://www.thevirtualbrain.org/tvb/zwei/neuroscience-publications
#
#

"""
An interactive Epileptor Model Visualiser.

Usage
::

    # Create and launch the interactive visualiser
    from tvb.simulator.plot.epileptor_plotter import EpileptorModelPlot
    ep = EpileptorModelPlot()
    ep.show()

"""

import numpy as np
import matplotlib.pyplot as plt

import ipywidgets as widgets
from IPython.display import display

from tvb.basic.neotraits.api import HasTraits, Attr
from tvb.simulator.lab import *
from tvb.simulator.models.epileptorcodim3 import EpileptorCodim3, EpileptorCodim3SlowMod

class EpileptorModelPlot(HasTraits):
    """
    The graphical interface for visualising the epileptor model, provide controls for setting:

        - select burster class of Epileptor [List]
        - select Simulation Length [slider]
        - select Epileptor Codim Slow Model [binary]
        - select dstar[binary]
        - specify dstar value [float]
        - select modification [binary]

    """

    conn = Attr(
        field_type=connectivity.Connectivity,
        label="Connectivity",
        default=connectivity.Connectivity.from_file(),
        doc=""" The surface connectvity to be used to simulate the model. """)
    
    coupling = Attr(
        field_type=coupling.Linear,
        label="Coupling",
        default=coupling.Linear(a=np.array([0.0152])),
        doc=""" The type of coupling to be used to simulate the model. """)

    integrator = Attr(
        field_type=integrators.HeunDeterministic,
        label="Integrator",
        default=integrators.HeunDeterministic(dt=2 ** -4),
        doc=""" The integrator to be used to simulate the model. """)

    monitors = Attr(
        field_type=tuple,
        label="Monitors",
        default=(monitors.TemporalAverage(period=2 ** -2),),
        doc=""" The tuple of monitors to be used to monitor the model. """)
    
    sim_length = Attr(
        field_type=int,
        label="Simulation Length",
        default=2**10,
        doc=""" The period for which the model should be simulated. """)

    
    def __init__(self, **kwargs):
        """ Initialise based on provided keywords or their traited defaults. """

        super(EpileptorModelPlot, self).__init__(**kwargs)

        self.burster_parameters = dict()
        self.default_burster_parameters = dict()
        self.default_widget_values = dict()

    def show(self, burster_class='default'):
        """ Generate the interactive Epileptor Model Simulator with the desired burster class. """

        self.set_default_burster_parameters()
        ui = self.create_ui(burster_class = burster_class)
        self.configure_model(slow=False)
        
        if burster_class == 'default':
            self.configure_sim()
        else:
            self.update_burster_parameters(burster_class)
        
        self.plot_model()
        display(ui)

    def create_ui(self, burster_class):
        """ Create the UI for the Interactive Plotter. """

        # Default Box Layout
        self.box_layout = widgets.Layout(border='solid 1px black',
                                        margin='0px 5px 5px 0px',
                                        padding='2px 2px 2px 2px')

        # Create desired UI and add Widgets
        self.add_burster_widgets(burster_class = burster_class)
        self.add_sim_widgets()
        self.add_model_widgets()
        
        # Final control widget box
        self.control_box = widgets.HBox([self.burster_param_box, self.sim_param_box ,self.model_param_box], layout=self.box_layout)
        
        # Output Widget
        self.fig = None
        self.op = widgets.Output()
        self.op.layout = self.box_layout

        # Widget and Output Grid 
        grid = widgets.GridBox([self.control_box, self.op], layout={'grid_template_rows':'225px 625px'})
        return grid

    def set_default_burster_parameters(self):
        """ Define default Bursters and their parameters. """

        self.default_burster_parameters['default'] = {'A':None, 'B':None, 'c':None}
        self.default_burster_parameters['c0'] = {'A':[0.2649, -0.05246, 0.2951], 'B':[0.2688, 0.05363, 0.2914], 'c':0.001, 'sl':11}
        self.default_burster_parameters['c2s'] = {'A':[0.3448,0.02285,0.2014], 'B':[0.3351,0.07465,0.2053], 'c':0.001, 'sl':10}
        self.default_burster_parameters['c3s'] = {'A':[0.2552,-0.0637,0.3014], 'B':[0.3496,0.0795,0.1774], 'c':0.0004, 'sl':13}
        self.default_burster_parameters['c10s'] = {'A':[0.3448,0.0228,0.2014], 'B':[0.3118,0.0670,0.2415], 'c':0.00005, 'sl':14}
        self.default_burster_parameters['c11s'] = {'A':[0.3131,-0.06743,0.2396], 'B':[0.3163,0.06846,0.2351], 'c':0.00004, 'sl':15}
        self.default_burster_parameters['c2b'] = {'A':[0.3216,0.0454,-0.2335], 'B':[0.285,0.05855,-0.2745], 'c':0.004, 'sl':10}
        self.default_burster_parameters['c4b'] = {'A':[0.1871,-0.02512,-0.3526], 'B':[0.2081,-0.01412,-0.3413], 'c':0.008, 'sl':10}
        self.default_burster_parameters['c14b'] = {'A':[0.3216,0.0454,-0.2335], 'B':[0.106,0.005238,-0.3857], 'c':0.002, 'sl':12}
        self.default_burster_parameters['c16b'] = {'A':[0.04098,-0.07373,-0.391], 'B':[-0.01301,-0.03242,-0.3985], 'c':0.004, 'sl':10}

    def add_burster_widgets(self, burster_class='default'):
        """ Add the Dropdown Widget to select desired Burster. """
        
        self.burster_choice_label = widgets.Label('Burster Choice:')
        self.burster_choice = widgets.Dropdown(options=list(self.default_burster_parameters.keys()), value=burster_class)

        self.buster_param_box_items = [self.burster_choice_label, self.burster_choice]
        self.burster_param_box = widgets.VBox(self.buster_param_box_items, layout=self.box_layout)

        self.default_widget_values['b_choice'] = self.burster_choice.value

        self.burster_choice.observe(self.update_burster_parameters, 'value')

    def add_sim_widgets(self):
        """ Add the slider for Simulation Length. """

        self.simulation_label = widgets.Label('Simulation Parameters')
        self.simulation_length_label = widgets.Label('Simulation Length: (in powers of 2)')
        self.simulation_length_slider = widgets.IntSlider(value=10, min=1, max=30, continuous_update=False)
        
        self.sim_param_box_items = [self.simulation_label, self.simulation_length_label, self.simulation_length_slider]
        self.sim_param_box = widgets.VBox(self.sim_param_box_items, layout=self.box_layout)

        self.default_widget_values['s_l'] = self.simulation_length_slider.value

        self.simulation_length_slider.observe(self.update_sim_parameters, 'value')

    def add_model_widgets(self):
        """ Add widgets to specify Model Parameters. """

        self.model_parameter_label = widgets.Label('Model Parameters')
        self.slowmod_checkbox = widgets.Checkbox(description='Slow Transition', value=False)
        self.dstar_checkbox = widgets.Checkbox(description='dstar', value=False)
        self.dstar_value = widgets.FloatText(description='dstar', value=0.0, continuous_update=False)
        self.mod_checkbox = widgets.Checkbox(description='Modification', value=False)
        
        self.model_param_items = [self.model_parameter_label,self.slowmod_checkbox,self.dstar_checkbox,self.dstar_value, self.mod_checkbox]
        self.model_param_box = widgets.VBox(self.model_param_items, layout=self.box_layout)

        self.slowmod_checkbox.observe(self.update_to_slow_model,'value')
        self.dstar_checkbox.observe(self.update_model_parameters,'value')
        self.dstar_value.observe(self.update_model_parameters,'value')
        self.mod_checkbox.observe(self.update_model_parameters,'value')

    def add_new_burster_class(self, burster_name=None,
                             burster_value_A = None, burster_value_B = None,
                             burster_value_c = None, burster_sl=10):
        """ Function to add a new desired burster class with specific parameters. """

        if burster_name is not None and burster_value_A is not None and \
           burster_value_B is not None and burster_value_c is not None:
            self.default_burster_parameters[burster_name] = {'A':burster_value_A, 'B':burster_value_B, 'c':burster_value_c, 'sl': burster_sl}
            self.burster_choice = widgets.Dropdown(options=list(self.default_burster_parameters.keys()), value='default')

    def update_to_slow_model(self, val):
        """ Toggle Model to Epileptor Codim Slow Model. """

        if self.slowmod_checkbox.value:
            self.burster_choice.value = 'default'
            self.burster_choice.disabled = True
            self.configure_model(slow=True)

            self.simulation_length_slider.value = 14
        else:
            self.burster_choice.disabled = False
            self.configure_model(slow=False)
            self.configure_sim()
            self.plot_model()

    def update_sim_parameters(self, val):
        """ Update Simulation Length and Simulate Model. """

        self.sim_length = 2**int(self.simulation_length_slider.value)
        self.configure_sim()
        if self.slow_model:
            self.plot_slow_model()
        else:
            self.plot_model()

    def update_burster_parameters(self, val):
        """ Update the Burster Parameters of Model, configure and simulate model. """

        self.burster_parameters = self.default_burster_parameters[self.burster_choice.value]

        A = self.burster_parameters['A']
        B = self.burster_parameters['B']
        c = self.burster_parameters['c']
        
        if self.burster_choice.value != 'default':
            self.model.mu1_start=np.array([-A[1]]) 
            self.model.mu2_start=np.array([A[0]]) 
            self.model.nu_start=np.array([A[2]])
            self.model.mu1_stop=np.array([-B[1]]) 
            self.model.mu2_stop=np.array([B[0]]) 
            self.model.nu_stop=np.array([B[2]]) 
            self.model.c=np.array([c])
            self.simulation_length_slider.value = int(self.burster_parameters['sl'])
        else:
            self.simulation_length_slider.value = 10
            self.configure_model(slow=False)
        
        self.configure_sim()
        self.plot_model()

    def update_model_parameters(self, val):
        """ Update Model Parameters, configure and simulate Model. """

        if self.dstar_checkbox.value or self.mod_checkbox.value:
            self.burster_choice.disabled = True
            self.model.dstar = np.array([self.dstar_value.value])
            self.model.modification = np.array([self.mod_checkbox.value])
            self.configure_sim()
            self.plot_model()
        else:
            self.burster_choice.disabled = False
            self.configure_model()
            self.configure_sim()
            self.plot_model()


    def configure_model(self, slow=False):
        """ Configure Model according to desired choice. (Normal/Slow) """

        if slow:
            self.slow_model = True
            self.model = EpileptorCodim3SlowMod()
        else:
            self.slow_model = False
            self.model = EpileptorCodim3(variables_of_interest=['x', 'y', 'z'])

    def configure_sim(self):
        """ Configure Simulation Parameters and run Simulation. """

        self.sim = simulator.Simulator(
            model=self.model,
            connectivity=self.conn,
            coupling=self.coupling,
            integrator=self.integrator,
            monitors=self.monitors,
            simulation_length=self.sim_length)
        
        self.sim = self.sim.configure()

        (self.tavg_time, self.tavg_data), = self.sim.run()

    def plot_model(self):
        """ Plot Normal Model Simulation Output. """
        self.op.clear_output(wait=True)
        with plt.ioff():
            if not self.fig:
                self.fig = plt.figure(figsize=(9,5))
            self.fig.clear()
            self.fig1 = self.fig.add_subplot(121)
            self.fig2 = self.fig.add_subplot(122, projection='3d')

            self.fig1.plot(self.tavg_time, self.tavg_data[:, 0, 0, 0], label='x')
            self.fig1.plot(self.tavg_time, self.tavg_data[:, 2, 0, 0], label='z')
            self.fig1.legend()
            self.fig1.grid(True)
            self.fig1.set_xlabel('Time (ms)')
            self.fig1.set_ylabel("Temporal average")

            self.fig2.plot(self.tavg_data[:, 0, 0, 0], self.tavg_data[:, 1, 0, 0], self.tavg_data[:, 2, 0, 0])

            plt.grid(True)
            plt.xlabel('x')
            plt.ylabel('y')
            with self.op:
                display(self.fig.canvas)

    def plot_slow_model(self):
        """ Plot Slow Model Simulation Output. """

        self.fig.clear()
        
        plt.plot(self.tavg_time, self.tavg_data[:, 0, 0, 0], label='x')
        plt.plot(self.tavg_time, self.tavg_data[:, 1, 0, 0], label='z')
        plt.legend()
        plt.grid(True)
        plt.xlabel('Time (ms)')
        plt.ylabel("Temporal average")